"""Template model for crystal problem for learning and name unification.
Now, we don't suggest using this class."""
import warnings
from abc import abstractmethod

import ase.data as ase_data
import numpy as np
import torch
import torch.nn.functional as F
from torch import Tensor, nn
from torch.nn import Embedding, Linear, LayerNorm, ModuleList, Softplus, ReLU, Sequential
from torch.nn import Module
from torch_scatter import segment_csr

from hpc2ml.nn.general import get_ptr


class TemplateCrystalModel(Module):
    """
    Template model for crystal problem for learning and name unification. Deprecation !

    The subclass should complete the `get_interactions_layer` and could redefined `get_readout_layer` function.

    Examples::
        def get_interactions_layer(self):
            self.layer_interaction = YourNet()

    Examples::
        def get_readout_layer(self):
            self.readout_layer = torch.nn.Sequential(...)

    Or re-write the `_forward` function。

    where the interactions are for center and neighbors interaction.
    where the readout_layer are for atomic properties merge into a whole.
    """

    def __init__(self,
                 nfeat_node=1,
                 nfeat_edge=3,
                 nfeat_state=0,

                 nc_node_hidden=16,
                 nc_edge_hidden=None,
                 nc_state_hidden=None,

                 edge_method=None,
                 node_method="z",
                 state_method=None,

                 nc_node_interaction=128,
                 num_interactions=1,

                 nnvocal=120,
                 nevocal=None,
                 nsvocal=None,

                 cutoff=10.0,
                 out_size=1,
                 readout='mean',

                 layer_interaction=None,
                 layer_readout=None,

                 add_node=None,
                 add_state=False,

                 norm=False,
                 mean=None,
                 std=None,
                 scale=None,
                 is_classification=False,
                 **kwargs
                 ):
        """
        Model for crystal problem.

        Args:
            nfeat_node: (int) input number of node feature (atom feature).
            nfeat_edge: (int) input number of bond feature. if ``num_edge_gaussians` offered,
            this parameter is neglect.
            nfeat_state: (int) input number of state feature.
            nnvocal: (int) number of atom, For generate the initial embedding matrix to on.
            nevocal: (int) number of bond types if bond attributes are types
            ngvocal: (int) number of global types if global attributes are types.
            behalf of node feature.
            nc_node_hidden: (int) nc_node_hidden for node feature.
            nc_node_interaction: (int) channels for node feature.
            num_interactions: (int) conv number.
            cutoff: (float) cutoff for calculate neighbor bond.
            readout: (str) Merge node method. such as "add","mean","max","mean".
            mean: (float) mean for y/energy.
            std: (float) std for y/energy.
            norm:(bool) False or True norm for y/energy.
            add_node: (torch.tensor shape (120,1)) properties for atom. such as target y is volumes of compound,
                add_node could be the atom volumes of all atom (H,H,He,Li,...). And you could copy the first term to
                make sure the `H` index start form 1.
            node_method: (bool,str) just used "z" or used "x" to calculate.
            layer_interaction: (Callable) torch module for interactions dynamically: pass the torch module to
            interactions parameter.static: re-define the ``get_interactions_layer`` and keep this parameter is None.
                the forward input is (h, edge_index, edge_weight, edge_attr, data=data)
            layer_readout: (Callable) torch module for interactions  dynamically: pass the torch module to
            interactions parameter. static: re-define the ``get_interactions_layer`` and keep this parameter is None.
            The forward input is (out,)
            add_state: (bool) add state attribute before output.
            is_classification: (bool) default False. 
            out_size:(int) number of out size. for regression,is 1 and for classification should be defined.
        """
        super(TemplateCrystalModel, self).__init__()

        warnings.warn("Template model for crystal problem for learning and name unification. "
                      "Now, we don't suggest using this class.", DeprecationWarning)

        self.interaction_kwargs = {}
        self.readout_kwargs = {}
        for k, v in kwargs.items():
            if "interaction_kwargs_" in k:
                self.interaction_kwargs[k.replace("interaction_kwargs_", "")] = v
            elif "readout_kwargs_" in k:
                self.readout_kwargs[k.replace("readout_kwargs_", "")] = v
            else:
                setattr(self, k, v)

        assert readout in ['add', 'sum', 'min', 'mean', "max"]

        if nnvocal < 120:
            print("Default, nvocal>=120, if you want simple the net work and "
                  "This network does not apply to other elements, the nvocal could be less but large than "
                  "the element type number in your data.")

        self.nnvocal = nnvocal
        self.nevocal = nevocal
        self.nsvocal = nsvocal

        self.nc_node_hidden = nc_node_hidden
        self.nc_state_hidden = nc_state_hidden
        self.nc_edge_hidden = nc_edge_hidden

        self.nfeat_state = nfeat_state
        self.nfeat_edge = nfeat_edge
        self.nfeat_node = nfeat_node

        self.node_method = node_method
        self.edge_method = edge_method
        self.state_method = state_method

        self.nc_node_interaction = nc_node_interaction
        self.num_interactions = num_interactions

        self.cutoff = cutoff
        self.readout = readout

        self.layer_interaction = layer_interaction
        self.layer_readout = layer_readout

        self.out_size = out_size

        self.norm = norm
        self.mean = mean
        self.std = std
        self.scale = scale
        self.add_state = add_state
        self.add_node = add_node
        self.is_classification = is_classification

        # 1. node (atom)
        if self.node_method == "z":
            if self.nfeat_node != 0:
                print("node_data=='z' would not use your self-defined 'x' data, "
                      "but element number Z.")
            self.layer_emb_node = Embedding(self.nnvocal, self.nc_node_hidden)
        elif self.node_method == "x":
            self.layer_emb_node = Sequential(Linear(self.nfeat_node, self.nc_node_hidden),
                                             Softplus(),
                                             Linear(self.nc_node_hidden, self.nc_node_hidden))
        else:
            raise ValueError("The node_data just accept 'z', 'x'.")

        # 2. edge
        if self.edge_method == "gaussian":
            assert self.nc_edge_hidden
            centers = kwargs.get("centers", None)
            self.layer_expand_edge = GaussianSmearing(0.0, self.cutoff, self.nc_edge_hidden)
        elif self.edge_method == "linear":
            assert self.nc_edge_hidden
            self.layer_expand_edge = Sequential(
                Linear(self.nfeat_edge, self.nc_edge_hidden),
                ReLU(),
                Linear(self.nc_edge_hidden, self.nc_edge_hidden), )
        elif self.edge_method == "embed":
            assert self.nevocal
            assert self.nc_edge_hidden
            self.layer_expand_edge = Sequential(
                Linear(self.nfeat_edge, self.nc_edge_hidden),
                ReLU(),
                Linear(self.nc_edge_hidden, self.nc_edge_hidden), )
        else:
            self.nc_edge_hidden = self.nfeat_edge

        # 3. state
        if self.nfeat_state is None and self.state_method == "embed":
            assert self.nevocal
            assert self.nc_state_hidden is not None
            # global state inputs are embedding integers
            self.layer_emb_state = Embedding(self.nsvocal, self.nc_state_hidden)

        else:
            self.nc_state_hidden = self.nfeat_state

        # 交互层 需要自定义 get_interactions_layer
        if layer_interaction is None:
            self.get_interactions_layer()
        elif isinstance(layer_interaction, ModuleList):
            self.SimpleGeoResNet(layer_interaction)
        elif isinstance(layer_interaction, Module):
            self.layer_interaction = layer_interaction
        else:
            raise NotImplementedError("Please implement get_interactions_layer function, "
                                      "or pass layer_interaction parameters.")
        # 合并层 需要自定义
        if layer_readout is None:
            self.get_readout_layer()
        elif isinstance(layer_readout, Module):
            self.layer_readout = layer_readout
        else:
            raise NotImplementedError("please implement get_readout_layer function, "
                                      "or pass layer_readout parameters.")

        # 原子性质嵌入
        if add_node is not None:
            self.add_node_layer = EmbAtomsProp(add_node)

        if self.add_state:
            assert self.nfeat_state > 0
            self.add_state_layer = Sequential(
                LayerNorm(self.nfeat_state),
                ReLU(),
                Linear(self.nfeat_state, 2 * self.nfeat_state),
                ReLU(),
                Linear(2 * self.nfeat_state, 2 * self.nfeat_state),
                ReLU(),
                Linear(2 * self.nfeat_state, self.out_size),
            )

        self.reset_parameters()

    def _forward(self, h, data):
        """This function can be re-write."""

        h = self.layer_interaction(h, data=data)

        if self.add_node is not None:
            h = h + self.add_node_layer(data.z)

        h = self.layer_readout(h, data)

        if self.add_state:
            assert hasattr(data, "state_attr"), "the ``add_state`` must accept ``state_attr`` in data."
            sta = self.add_state_layer(data.state_attr)
            h = h + sta

        return h

    def forward(self, data):

        assert hasattr(data, "z")
        assert hasattr(data, "pos")
        assert hasattr(data, "batch")
        z = data.z

        # 原子处理数据阶段
        if self.node_method == "z":
            assert z.dim() == 1 and z.dtype == torch.long
            h = self.layer_emb_node(z)
        elif self.node_method == "x":
            assert hasattr(data, "x")
            h = self.layer_emb_node(data.x)
        else:
            raise NotImplementedError

        data.x = h

        # 键处理数据阶段
        if self.nc_edge_hidden:
            if self.edge_method == "linear":
                edge_attr = self.layer_expand_edge(data.edge_attr)
                data.edge_attr = edge_attr
            elif self.edge_method == "embed":
                assert data.edge_weight.dtype == torch.int64
                edge_attr = self.layer_expand_edge(data.edge_weight)
                data.edge_attr = edge_attr
            elif self.edge_method == "gaussian":
                edge_attr = self.layer_expand_edge(data.edge_weight)
                data.edge_attr = edge_attr
            else:
                pass

        if hasattr(data, "state_attr"):
            if self.nfeat_state is None and self.state_method == "embed":
                assert data.edge_weight.dtype == torch.int64
                data.edge_attr = self.layer_emb_state(data.edge_attr)
            else:
                pass

        h = self._forward(h, data)  # key function.

        h = self.output_forward(h)

        return h

    @abstractmethod
    def get_interactions_layer(self):
        """This part shloud re-defined. And must add the ``interactions`` attribute.

        Examples::

            def get_interactions_layer(self):
                self.layer_interaction = YourNet()

        """

    def get_readout_layer(self):
        """This part should re-defined. And must add the ``readout_layer`` attribute.

        Examples::
            def get_readout_layer(self):
                self.readout_layer = torch.nn.Sequential(...)

        Examples::

            def get_readout_layer(self):
                self.readout_layer = YourNet()

        """
        if "readout_kwargs_layers_size" in self.readout_kwargs:
            self.layer_readout = GeneralReadOutLayer(**self.readout_kwargs)
        else:
            if self.is_classification:
                last_layer = nn.Sigmoid
            else:
                last_layer = None
            self.layer_readout = GeneralReadOutLayer(
                [self.nc_node_interaction, self.readout, 2 * self.nc_node_interaction,
                 2 * self.nc_node_interaction, self.out_size], last_layer=last_layer, **self.readout_kwargs)

    def output_forward(self, out):
        """Last dispose."""
        if self.mean is not None and self.std is not None:
            out = out * self.std + self.mean
        if self.norm is True:
            out = torch.norm(out, dim=-1, keepdim=True)
        if self.scale is not None:
            out = self.scale * out
        return out

    def forward_weight_attr(self, data):
        """Add edge_weight or edge_attr by another, use mannully."""

        if not hasattr(data, "edge_weight") or data.edge_weight is None:
            if not hasattr(data, "edge_attr") or data.edge_attr is None:
                raise NotImplementedError("Must offer edge_weight or edge_attr.")
            else:
                if data.edge_attr.shape[1] == 1:
                    data.edge_weight = data.edge_attr.reshape(-1)
                else:
                    data.edge_weight = torch.norm(data.edge_attr, dim=1, keepdim=True)

        if not hasattr(data, "edge_attr") or data.edge_attr is None:
            if not hasattr(data, "edge_weight") or data.edge_weight is None:
                raise NotImplementedError("Must offer edge_weight or edge_attr.")
            else:
                data.edge_attr = data.edge_weight.reshape(-1, 1)

        return data

    def __repr__(self):
        return (f'{self.__class__.__name__}('
                f'nc_node_hidden={self.nc_node_hidden}, '
                f'nc_edge_hidden={self.nc_edge_hidden}, '
                f'nc_state_hidden={self.nc_state_hidden}, '
                f'nc_node_interaction={self.nc_node_interaction}, '
                f'num_interactions={self.num_interactions}, '
                f'cutoff={self.cutoff})')

    def reset_parameters(self):
        if hasattr(self, "layer_emb_node"):
            try:
                self.layer_emb_node.reset_parameters()
            except AttributeError:
                pass
        if hasattr(self, "layer_expand_edge"):
            try:
                self.layer_expand_edge.reset_parameters()
            except AttributeError:
                pass
        if hasattr(self, "layer_emb_state"):
            try:
                self.layer_emb_state.reset_parameters()
            except AttributeError:
                pass
        if hasattr(self, "layer_interaction"):
            try:
                self.layer_interaction.reset_parameters()
            except AttributeError:
                pass
        if hasattr(self, "layer_readout"):
            try:
                self.layer_readout.reset_parameters()
            except AttributeError:
                pass


class _ReadOutLayer(Module):
    """Merge node."""

    def __init__(self, channels, out_size=1, readout="add"):
        super(_ReadOutLayer, self).__init__()
        self.readout = readout
        self.ro_lin1 = Linear(channels, channels * 10)
        self.ro_lrelu1 = nn.LeakyReLU()
        self.ro_lin2 = Linear(channels * 10, channels * 5)
        self.ro_lrelu2 = nn.LeakyReLU()
        self.ro_lin3 = Linear(channels * 5, out_size)

    def forward(self, h, data):
        batch = data.batch
        h = self.ro_lin1(h)
        h = self.ro_lrelu1(h)
        h = segment_csr(h, get_ptr(batch), reduce=self.readout)
        h = self.ro_lin2(h)
        h = self.ro_lrelu2(h)
        h = self.ro_lin3(h)
        return h


class EmbAtomsProp(Module):
    """Embedding of atomic properties."""

    def __init__(self, array="atomic_radii"):

        super().__init__()

        if array == "atomic_mass":
            array = torch.from_numpy(ase_data.atomic_masses)  # 嵌入原子质量
        elif array == "atomic_radii":
            array = torch.from_numpy(ase_data.covalent_radii)  # 嵌入共价半径
        elif isinstance(array, np.ndarray):
            assert array.shape[0] == 120
            array = torch.from_numpy(array)
        elif isinstance(array, Tensor):
            assert array.shape[0] == 120
        else:
            raise NotImplementedError("just accept str,np,ndarray or tensor with shape (120,)")
        # 嵌入原子属性，需要的时候运行本函数
        # (嵌入别太多，容易慢，大多数情况下用不到。)
        # 缓冲buffer必须要登记注册才会有效,如果仅仅将张量赋值给Module模块的属性,不会被自动转为缓冲buffer.
        # 因而也无法被state_dict()、buffers()、named_buffers()访问到。
        self.register_buffer('atomic_temp', array)

    def forward(self, z):
        return self.atomic_temp[z]


class GeneralReadOutLayer(Module):
    """General Merge node.
    """

    def __init__(self, layers_size=(128, "add", 32, 1), last_layer=None, active_layer_type="Softplus"):
        super(GeneralReadOutLayer, self).__init__()
        l = len(layers_size)
        readout = [i for i in layers_size if isinstance(i, str)]
        if active_layer_type == "Softplus":
            active_layer = nn.Softplus
        elif active_layer_type == "SoftplusShift":
            active_layer = ShiftedSoftplus
        elif active_layer_type == "ReLU":
            active_layer = nn.ReLU
        elif active_layer_type == "LeakyReLU":
            active_layer = nn.LeakyReLU
        elif isinstance(active_layer_type, Module):
            active_layer = active_layer_type.__class__
        else:
            raise NotImplementedError("Can't identify the type of layer.")

        assert len(readout) == 1, "The readout layer must be set one time, please there are {} layer: {}.".format(
            len(readout), readout)

        readout_site = [n for n, i in enumerate(layers_size) if isinstance(i, str)][0]
        readout = layers_size[readout_site]
        assert readout in ('sum', 'max', 'min', 'mean', 'add')
        self.readout = readout
        layers_size = list(layers_size)
        layers_size[readout_site] = layers_size[readout_site - 1]

        part_first = []
        part_second = []
        i = 0
        while i < l - 1:
            if i < readout_site - 1:
                part_first.append(Linear(layers_size[i], layers_size[i + 1]))
                part_first.append(active_layer())
            elif i < readout_site:
                pass
            else:
                part_second.append(active_layer())
                part_second.append(Linear(layers_size[i], layers_size[i + 1]))

            i += 1

        self.ro_seq1 = Sequential(*part_first)
        self.ro_seq2 = Sequential(*part_second)
        self.ro_last_layer = last_layer

    def reset_parameters(self):
        self.ro_seq1.reset_parameters()
        self.ro_seq2.reset_parameters()
        self.ro_last_layer.reset_parameters()

    def forward(self, h, data):
        batch = data.batch
        if len(self.ro_seq1) > 0:
            h = self.ro_seq1(h)
        h = segment_csr(h, get_ptr(batch), reduce=self.readout)
        if len(self.ro_seq2) > 0:
            h = self.ro_seq2(h)
        if self.ro_last_layer is not None:
            h = self.ro_last_layer(h)
        return h


class SimpleGeoResNet(Module):
    """Simple ResNet."""

    def __init__(self, module_list: ModuleList):
        super().__init__()
        self.res_list = module_list

    def forward(self, h, data):
        h0 = h
        for interaction in self.res_list:
            h = interaction(h0, data=data)
        return h


class ShiftedSoftplus(torch.nn.Module):
    def __init__(self):
        super().__init__()
        self.shift = torch.log(torch.tensor(2.0)).item()

    def forward(self, x):
        return F.softplus(x) - self.shift


class GaussianSmearing(torch.nn.Module):
    def __init__(self, start=0.0, stop=5.0, num_gaussian=50):
        super().__init__()
        offset = torch.linspace(start, stop, num_gaussian)
        self.coef = -0.5 / (offset[1] - offset[0]).item() ** 2
        self.register_buffer('offset', offset)

    def forward(self, dist):
        dist = dist.view(-1, 1) - self.offset.view(1, -1)
        return torch.exp(self.coef * torch.pow(dist, 2))


def get_active_layer(name: str):
    """Return active layer."""
    if name == "Softplus":
        active_layer = nn.Softplus
    elif name == "SoftplusShift":
        active_layer = ShiftedSoftplus
    elif name == "ReLU":
        active_layer = nn.ReLU
    elif name == "LeakyReLU":
        active_layer = nn.LeakyReLU
    else:
        raise NotImplementedError
    return active_layer



import itertools
from typing import Literal, Sequence

import numpy as np
from pymatgen.analysis.structure_matcher import StructureMatcher, ElementComparator, AbstractComparator
from pymatgen.core.composition import SpeciesLike
from tqdm import tqdm


class StructureMatcherExtend(StructureMatcher):
    """
    Class to match structures by similarity.

    Algorithm:

    1. Given two structures: s1 and s2
    2. Optional: Reduce to primitive cells.
    3. If the number of sites do not match, return False
    4. Reduce to s1 and s2 to Niggli Cells
    5. Optional: Scale s1 and s2 to same volume.
    6. Optional: Remove oxidation states associated with sites
    7. Find all possible lattice vectors for s2 within shell of ltol.
    8. For s1, translate an atom in the smallest set to the origin
    9. For s2: find all valid lattices from permutations of the list
       of lattice vectors (invalid if: det(Lattice Matrix) < half
       volume of original s2 lattice)
    10. For each valid lattice:

        a. If the lattice angles of are within tolerance of s1,
           basis change s2 into new lattice.
        b. For each atom in the smallest set of s2:

            i. Translate to origin and compare fractional sites in
            structure within a fractional tolerance.
            ii. If true:

                ia. Convert both lattices to Cartesian and place
                both structures on an average lattice
                ib. Compute and return the average and max rms
                displacement between the two structures normalized
                by the average free length per atom

                if fit function called:
                    if normalized max rms displacement is less than
                    stol. Return True

                if get_rms_dist function called:
                    if normalized average rms displacement is less
                    than the stored rms displacement, store and
                    continue. (This function will search all possible
                    lattices for the smallest average rms displacement
                    between the two structures)
    """

    def __init__(
            self,
            ltol: float = 0.2,
            stol: float = 0.3,
            angle_tol: float = 5,
            primitive_cell: bool = True,
            scale: bool = True,
            attempt_supercell: bool = False,
            allow_subset: bool = False,
            comparator: AbstractComparator = None,
            supercell_size: Literal["num_sites", "num_atoms", "volume"] = "num_sites",
            ignored_species: Sequence[SpeciesLike] = (),
    ):
        """
        Args:
            ltol (float): Fractional length tolerance. Default is 0.2.
            stol (float): Site tolerance. Defined as the fraction of the
                average free length per atom := ( V / Nsites ) ** (1/3)
                Default is 0.3.
            angle_tol (float): Angle tolerance in degrees. Default is 5 degrees.
            primitive_cell (bool): If true: input structures will be reduced to
                primitive cells prior to matching. Default to True.
            scale (bool): Input structures are scaled to equivalent volume if
               true; For exact matching, set to False.
            attempt_supercell (bool): If set to True and number of sites in
                cells differ after a primitive cell reduction (divisible by an
                integer) attempts to generate a supercell transformation of the
                smaller cell which is equivalent to the larger structure.
            allow_subset (bool): Allow one structure to match to the subset of
                another structure. Eg. Matching of an ordered structure onto a
                disordered one, or matching a delithiated to a lithiated
                structure. This option cannot be combined with
                attempt_supercell, or with structure grouping.
            comparator (Comparator): A comparator object implementing an equals
                method that declares equivalency of sites. Default is
                SpeciesComparator, which implies rigid species
                mapping, i.e., Fe2+ only matches Fe2+ and not Fe3+.

                Other comparators are provided, e.g., ElementComparator which
                matches only the elements and not the species.

                The reason why a comparator object is used instead of
                supplying a comparison function is that it is not possible to
                pickle a function, which makes it otherwise difficult to use
                StructureMatcher with Python's multiprocessing.
            supercell_size (str or list): Method to use for determining the
                size of a supercell (if applicable). Possible values are
                'num_sites', 'num_atoms', 'volume', or an element or list of elements
                present in both structures.
            ignored_species (list): A list of ions to be ignored in matching.
                Useful for matching structures that have similar frameworks
                except for certain ions, e.g., Li-ion intercalation frameworks.
                This is more useful than allow_subset because it allows better
                control over what species are ignored in the matching.
        """
        if comparator is None:
            comparator = ElementComparator()
        super(StructureMatcherExtend, self).__init__(ltol=ltol,
                                                     stol=stol, scale=scale,
                                                     angle_tol=angle_tol,
                                                     primitive_cell=primitive_cell,
                                                     attempt_supercell=attempt_supercell,
                                                     allow_subset=allow_subset,
                                                     comparator=comparator,
                                                     supercell_size=supercell_size,
                                                     ignored_species=ignored_species)

    def group_structures_and_index(self, s_list, anonymous=False):

        """
        Given a list of structures, use fit to group
        them by structural equality.

        Args:
            s_list ([Structure]): List of structures to be grouped
            anonymous (bool): Whether to use anonymous mode.

        Returns:
            A list of lists of matched structures
            Assumption: if s1 == s2 but s1 != s3, than s2 and s3 will be put
            in different groups without comparison.
        """
        if self._subset:
            raise ValueError("allow_subset cannot be used with group_structures")

        original_s_list = list(s_list)
        s_list = self._process_species(s_list)
        # Prepare reduced structures beforehand
        s_list = [self._get_reduced_structure(s, self._primitive_cell, niggli=True) for s in s_list]

        # Use structure hash to pre-group structures
        if anonymous:

            def c_hash(c):
                return c.anonymized_formula

        else:
            c_hash = self._comparator.get_hash

        def s_hash(s):
            return c_hash(s[1].composition)

        sorted_s_list = sorted(enumerate(s_list), key=s_hash)
        all_groups = []
        index_groups = []

        # For each pre-grouped list of structures, perform actual matching.
        for _, g in tqdm(itertools.groupby(sorted_s_list, key=s_hash)):
            unmatched = list(g)
            while len(unmatched) > 0:
                i, refs = unmatched.pop(0)
                matches = [i]
                if anonymous:
                    inds = filter(
                        lambda i: self.fit_anonymous(refs, unmatched[i][1], skip_structure_reduction=True),
                        list(range(len(unmatched))),
                    )
                else:
                    inds = filter(
                        lambda i: self.fit(refs, unmatched[i][1], skip_structure_reduction=True),
                        list(range(len(unmatched))),
                    )
                inds = list(inds)
                matches.extend([unmatched[i][0] for i in inds])
                unmatched = [unmatched[i] for i in range(len(unmatched)) if i not in inds]
                all_groups.append([original_s_list[i] for i in matches])
                index_groups.append(matches)

        return all_groups, index_groups

    def select_min_in_group(self, structure, array):
        """group the structures, and select the min index in each group according the values in array."""
        gp, index = self.group_structures_and_index(structure)
        index = [np.array(i) for i in index]
        select_index = []
        # log = {}
        for indexi in index:
            if len(indexi) == 1:
                ini = indexi[0]
            else:
                gp_e = array[indexi]
                min_ind = np.argmin(gp_e)
                ini = indexi[min_ind]
                # sel_k = gp_k[min_ind]
                # log.update({sel_k: (tuple(gp_k), tuple(gp_e))})
            select_index.append(ini)
        return np.array(select_index)



import warnings

import torch
from torch_geometric.transforms import BaseTransform


class GaussianSmearing(BaseTransform):
    """Smear the radius shape (num_node,1) to shape (num_node, num_edge_gaussians)."""

    def __init__(self, start=0.0, stop=5.0, num_edge_gaussians=50):
        super(GaussianSmearing, self).__init__()
        self.offset = torch.linspace(start, stop, num_edge_gaussians)
        self.coef = -0.5 / (self.offset[1] - self.offset[0]).item() ** 2

    def __call__(self, data):
        dist = data.edge_weight
        dist = dist.view(-1, 1) - self.offset.view(1, -1)
        if hasattr(data, "edge_attr") and data.edge_attr.shape[1] != 1:
            warnings.warn("The old edge_attr is covered by smearing edge_weight", UserWarning)
        data.edge_attr = torch.exp(self.coef * torch.pow(dist, 2))

        return data



import pandas as pd
from pymatgen.core import Structure

from hpc2ml.data.io import aaa
from hpc2ml.db.ase_db_extension.ext_db_csv import _decode_dct


def sparse_csv(csv_file: str, fmt=lambda x: x):
    """convert table."""
    if isinstance(csv_file, str):
        file = pd.read_csv(csv_file, index_col=0)
    else:
        file = csv_file

    data = file.to_dict(orient="index")

    dct = {}
    for k in data.keys():
        value = _decode_dct(data[k])
        if "atoms" in value:
            st = aaa.get_structure(value["atoms"])
            value["structure"] = st
        elif "structure" in value:
            if isinstance(value["structure"], Structure):
                pass
            else:
                try:
                    st = Structure.from_str(value["structure"], fmt="json")
                except:
                    st = fmt(value["structure"])
                value["structure"] = st
        else:
            raise NotImplementedError("Can't defined structure by names : `atoms` or `structure` in csv")
        dct = {k: value}

    return dct



"""
Class to holde a bunch of hyperparameters associate with either training or a model.

The interface is inteneded to be as close to the wandb.config class as possible. But it does not have any locked
entries as in wandb.config

Examples:

    Initialization
    ```
    config = Config()
    config = Config(dict(a=1, b=2))
    ```

    add a new parameter

    ```
    config['key'] = default_value
    config.key = default_value
    ```

    set up typehint for a parameter
    ```
    config['_key_type'] = int
    config._key_type = int
    config.set_type(key, int)
    ```

    update with a dictionary
    ```
    config.update(dictionary={'a':3, 'b':4})
    ```

    If a parameter is updated, the updated value will be formatted back to the same type.

"""
import inspect
from copy import deepcopy
from typing import Optional

from .savenload import save_file, load_file


class Config(object):
    def __init__(
            self,
            config: Optional[dict] = None,
            allow_list: Optional[list] = None,
            exclude_keys: Optional[list] = None,
    ):

        object.__setattr__(self, "_items", dict())
        object.__setattr__(self, "_item_types", dict())
        object.__setattr__(self, "_allow_list", list())
        object.__setattr__(self, "_allow_all", True)

        if allow_list is not None:
            self.add_allow_list(allow_list, default_values={})

        if config is not None and exclude_keys is not None:
            config = {
                key: value for key, value in config.items() if key not in exclude_keys
            }
        if config is not None:
            self.update(config)

    def __repr__(self):
        return str(dict(self))

    __str__ = __repr__

    def keys(self):
        return self._items.keys()

    def _as_dict(self):
        return self._items

    def as_dict(self):
        return dict(self)

    def __getitem__(self, key):
        return self._items[key]

    def get_type(self, key):
        """Get Typehint from item_types dict or previous defined value
        Args:

            key: name of the variable
        """

        return self._item_types.get(key, None)

    def set_type(self, key, typehint):
        """set typehint for a variable

        Args:

            key: name of the variable
            typehint: type of the variable
        """

        self._item_types[key] = typehint

    def add_allow_list(self, keys, default_values={}):
        """add key to allow_list"""

        object.__setattr__(self, "_allow_all", False)
        object.__setattr__(
            self, "_allow_list", list(set(self._allow_list).union(set(keys)))
        )
        self.update(default_values)

    def allow_list(self):
        return self._allow_list

    def __setitem__(self, key, val):

        # typehint
        if key.endswith("_type") and key.startswith("_"):

            k = key[1:-5]
            if (not self._allow_all) and key not in self._allow_list:
                return None

            self._item_types[k] = val

        # normal value
        else:

            if (not self._allow_all) and key not in self._allow_list:
                return None

            typehint = self.get_type(key)

            # try to format the variable
            try:
                val = typehint(val) if typehint is not None else val
            except Exception:
                raise TypeError(
                    f"Wrong Type: Parameter {key} should be {typehint} type."
                    f"But {type(val)} is given"
                )

            self._items[key] = deepcopy(val)
            return key

    def items(self):
        return self._items.items()

    __setattr__ = __setitem__

    def __getattr__(self, key):
        return self.__getitem__(key)

    def __contains__(self, key):
        return key in self._items

    def pop(self, *args):
        return self._items.pop(*args)

    def update_w_prefix(
            self,
            dictionary: dict,
            prefix: str,
            allow_val_change=None,
    ):
        """Mock of wandb.config function

        Add a dictionary of parameters to the
        The key of the parameter cannot be started as "_"

        Args:

            dictionary (dict): dictionary of parameters and their typehint to update
            allow_val_change (None): mock for wandb.config, not used.

        Returns:

        """

        # override with prefix
        l_prefix = len(prefix) + 1
        prefix_dict = {
            k[l_prefix:]: v for k, v in dictionary.items() if k.startswith(prefix + "_")
        }
        keys = self.update(prefix_dict, allow_val_change=allow_val_change)
        keys = {k: f"{prefix}_{k}" for k in keys}

        for suffix in ["params", "kwargs"]:
            if f"{prefix}_{suffix}" in dictionary:
                key3 = self.update(
                    dictionary[f"{prefix}_{suffix}"],
                    allow_val_change=allow_val_change,
                )
                keys.update({k: f"{prefix}_{suffix}.{k}" for k in key3})
        return keys

    def update(self, dictionary: dict, allow_val_change=None):
        """Mock of wandb.config function

        Add a dictionary of parameters to the config
        The key of the parameter cannot be started as "_"

        Args:

            dictionary (dict): dictionary of parameters and their typehint to update
            allow_val_change (None): mock for wandb.config, not used.

        Returns:
            keys (set): set of keys being udpated

        """

        keys = []

        # first log in all typehints or hidden variables
        for k, value in dictionary.items():
            if k.startswith("_"):
                keys += [self.__setitem__(k, value)]

        # then log in the values
        for k, value in dictionary.items():
            if not k.startswith("_"):
                keys += [self.__setitem__(k, value)]

        return set(keys) - set([None])

    def get(self, *args):
        return self._items.get(*args)

    def persist(self):
        """mock wandb.config function"""
        pass

    def setdefaults(self, d):
        """mock wandb.config function"""
        pass

    def update_locked(self, d, user=None):
        """mock wandb.config function"""
        pass

    def save(self, filename: str, format: Optional[str] = None):
        """Print config to file."""

        supported_formats = {"yaml": ("yml", "yaml"), "json": "json"}
        return save_file(
            item=dict(self),
            supported_formats=supported_formats,
            filename=filename,
            enforced_format=format,
        )

    @staticmethod
    def from_file(filename: str, format: Optional[str] = None, defaults: dict = {}):
        """Load arguments from file"""

        supported_formats = {"yaml": ("yml", "yaml"), "json": "json"}
        dictionary = load_file(
            supported_formats=supported_formats,
            filename=filename,
            enforced_format=format,
        )
        return Config.from_dict(dictionary, defaults)

    @staticmethod
    def from_dict(dictionary: dict, defaults: dict = {}):
        c = Config(defaults)
        c.update(dictionary)
        return c

    @staticmethod
    def from_class(class_type, remove_kwargs: bool = False):
        """return Config class instance based on init function of the input class
        the instance will only allow to store init function related variables
        the type hints are all set to None, so no automatic format conversion is applied

        class_type: torch.module children class type, i.e. .nequip.Nequip
        remove_kwargs (optional, bool): the same as Config.from_function

        Returns:

        config (Config):
        """

        if inspect.isclass(class_type):
            return Config.from_function(
                class_type.__init__, remove_kwargs=remove_kwargs
            )
        elif callable(class_type):
            return Config.from_function(class_type, remove_kwargs=remove_kwargs)
        else:
            raise ValueError(
                f"from_class only takes class type or callable, but got {class_type}"
            )

    @staticmethod
    def from_function(function, remove_kwargs=False):
        """return Config class instance based on the function of the input class
        the instance will only allow to store init function related variables
        the type hints are all set to None, so no automatic format conversion is applied

        Args:

        function: function name
        remove_kwargs (optional, bool): if True, kwargs are removed from the keys
             and the returned instance will only takes the init params of the class_type.
             if False and kwargs exists, the config only initialized with the default param values,
             but it can take any other keys

        Returns:

        config (Config):
        """

        sig = inspect.signature(function)

        default_params = {
            k: v.default
            for k, v in sig.parameters.items()
            if v.default is not inspect.Parameter.empty
        }
        param_keys = list(sig.parameters.keys())
        if param_keys[0] == "self":
            param_keys = param_keys[1:]

        for key in param_keys:
            default_params[f"_{key}_type"] = None

        # do not restrict variables when kwargs exists
        if "kwargs" in param_keys and not remove_kwargs:
            return Config(config=default_params)
        elif "kwargs" in param_keys:
            param_keys.remove("kwargs")
            return Config(config=default_params, allow_list=param_keys)
        else:
            return Config(config=default_params, allow_list=param_keys)

    load = from_file



# -*- coding: utf-8 -*-
# @Time  : 2022/10/18 23:06
# @Author : boliqq07
# @Software: PyCharm
# @License: MIT License

from typing import Iterator, Optional, Sized

import numpy as np
import torch
from sklearn.model_selection import train_test_split
from torch.utils.data import Sampler


class RandomGroupSampler(Sampler[int]):
    r"""Samples elements in group randomly.
    If without replacement, then sample from a shuffled dataset.
    If with replacement, then user can specify :attr:`num_samples` to draw.

    Args:
        data_source (Dataset): dataset to sample from
        replacement (bool): samples are drawn on-demand with replacement if ``True``, default=``False``
        num_samples (int): number of samples to draw, default=`len(dataset)`.
        generator (Generator): Generator used in sampling.
        group (int): group size.
    """
    data_source: Sized
    replacement: bool

    def __init__(self, data_source: Sized, replacement: bool = False,
                 num_samples: Optional[int] = None, generator=None, group=2) -> None:
        self.data_source = data_source
        self.replacement = replacement
        self._num_samples = num_samples  # old
        self.generator = generator
        self.group = group
        super(RandomGroupSampler, self).__init__(data_source=data_source)

        if not isinstance(self.replacement, bool):
            raise TypeError("replacement should be a boolean value, but got "
                            "replacement={}".format(self.replacement))

        if not isinstance(self.num_samples, int) or self.num_samples <= 0:
            raise ValueError("num_samples should be a positive integer "
                             "value, but got num_samples={}".format(self.num_samples))

    @property
    def num_samples(self) -> int:  # half
        # dataset size might change at runtime
        if self._num_samples is None:
            assert len(self.data_source) % self.group == 0, f"Just For data size can be divided evenly by {self.group}."
            return len(self.data_source) // 2
        else:
            assert self._num_samples % self.group == 0, f"Just For data size can be divided evenly by {self.group}."
            return self._num_samples // 2

    def __iter__(self) -> Iterator[int]:

        def back(arr: torch.Tensor):
            arrs = [torch.sub((arr + 1) * self.group, self.group - i) for i in range(self.group)]
            arrs = torch.vstack(arrs).T
            arr_all = torch.ravel(arrs)
            return arr_all

        n = len(self.data_source) // 2
        if self.generator is None:
            seed = int(torch.empty((), dtype=torch.int64).random_().item())
            generator = torch.Generator()
            generator.manual_seed(seed)
        else:
            generator = self.generator

        if self.replacement:
            for _ in range(self.num_samples // 32):
                yield from back(torch.randint(high=n, size=(32,), dtype=torch.int64, generator=generator)).tolist()
            yield from back(
                torch.randint(high=n, size=(self.num_samples % 32,), dtype=torch.int64, generator=generator)).tolist()
        else:
            for _ in range(self.num_samples // n):
                yield from back(torch.randperm(n, generator=generator)).tolist()
            yield from back(torch.randperm(n, generator=generator)).tolist()[:self.num_samples % n]

    def __len__(self) -> int:  # half
        return self.num_samples


def train_test_split_2_group(dataset, train_size=0.80, test_size=0.20, shuffle=True, random_state=None):
    """keep group 2 in old rank."""
    train_index_raw, test_index_raw = train_test_split(np.arange(len(dataset) // 2), train_size=train_size,
                                                       test_size=test_size, shuffle=shuffle, random_state=random_state)

    train_index1 = 2 * train_index_raw
    test_index1 = 2 * test_index_raw

    train_index2 = 2 * train_index_raw + 1
    test_index2 = 2 * test_index_raw + 1

    train_index = np.vstack([train_index1, train_index2]).T.ravel()
    test_index = np.vstack([test_index1, test_index2]).T.ravel()

    train_dataset = dataset[train_index.tolist()]
    test_dataset = dataset[test_index.tolist()]
    return train_dataset, test_dataset



import torch
import torch.nn.functional
from e3nn.o3 import Irreps
from e3nn.util.jit import compile_mode

from hpc2ml.data.nequlip_support import AtomicDataDict
from hpc2ml.data.nequlip_support.AtomicDataDict import ST
from .._graph_mixin import GraphModuleMixin


@compile_mode("script")
class OneHotAtomEncoding(GraphModuleMixin, torch.nn.Module):
    """Copmute a one-hot floating point encoding of atoms' discrete atom types.
    out_field in [AtomicDataDict.NODE_ATTRS_KEY,AtomicDataDict.NODE_FEATURES_KEY,]

    Args:
        copy_features: If ``True`` (default), copy out_field to another.
    """

    num_types: int
    copy_features: bool
    out_field: str
    another_field: str

    def __init__(
            self,
            irreps_in=None,
            out_field=AtomicDataDict.NODE_ATTRS_KEY,
            num_types: int = 50,
            copy_features: bool = True,

    ):
        super().__init__()
        self.num_types = num_types
        self.copy_features = copy_features
        self.out_field = out_field

        another_field = [i for i in [AtomicDataDict.NODE_FEATURES_KEY,
                                     AtomicDataDict.NODE_ATTRS_KEY] if i != out_field][0]
        self.another_field = another_field

        # Output irreps are num_types even (invariant) scalars nx0e
        irreps_out = {self.out_field: Irreps([(self.num_types, (0, 1))])}

        if self.copy_features:
            irreps_out[self.another_field] = irreps_out[self.out_field]
        self._init_irreps(irreps_in=irreps_in, irreps_out=irreps_out)

    def forward(self, data: ST) -> ST:
        type_numbers = data[AtomicDataDict.ATOM_TYPE_KEY].squeeze(-1)
        one_hot = torch.nn.functional.one_hot(
            type_numbers, num_classes=self.num_types
        ).to(device=type_numbers.device, dtype=data[AtomicDataDict.POSITIONS_KEY].dtype)
        data[self.out_field] = one_hot
        if self.copy_features:
            data[self.another_field] = one_hot
        return data



from typing import Union

import torch
from e3nn import o3
from torch_geometric.data import Batch, Data

from hpc2ml.data.nequlip_support import AtomicDataDict
from hpc2ml.data.batchdata import to_atomicdatadict
from hpc2ml.nn.nequlipnn.nn import AtomwiseLinear, ConvNetLayer, InteractionBlock, AtomwiseReduce
from hpc2ml.nn.nequlipnn.nn.cutoffs import PolynomialCutoff
from hpc2ml.nn.nequlipnn.nn.embedding import OneHotAtomEncoding, SphericalHarmonicEdgeAttrs, \
    RadialBasisEdgeEncoding
from hpc2ml.nn.nequlipnn.nn.radial_basis import BesselBasis
from nequip.nn import GraphModuleMixin


def config(lmax=1, parity=True, num_features=16, **kw):
    kw_ = {"readout": "sum",
           "r_max": 4.0,
           "num_layers": 3,
           "num_basis": 8,
           "num_types": 80,
           "PolynomialCutoff_p": 6,
           "nonlinearity_type": "gate"}

    kw.update({"chemical_embedding_irreps_out": repr(o3.Irreps([(num_features, (0, 1))]))})  # n scalars

    kw.update({"irreps_edge_sh": repr(o3.Irreps.spherical_harmonics(lmax=lmax, p=-1 if parity else 1))})

    kw.update({"feature_irreps_hidden": repr(o3.Irreps(
        [(num_features, (l, p))
         for p in ((1, -1) if parity else (1,))
         for l in range(lmax + 1)
         ]))})

    kw.update({"conv_to_output_hidden_irreps_out": repr(o3.Irreps([(max(1, num_features // 2), (0, 1))]))})
    kw_.update(kw)
    return kw_


class NeqlipNN(GraphModuleMixin, torch.nn.Module):
    def __init__(self, lmax=1, parity=True, num_features=32,
                 ):
        kw = config(lmax=lmax, parity=parity, num_features=num_features, )
        num_types = kw["num_types"]
        irreps_edge_sh = kw["irreps_edge_sh"]
        num_basis = kw["num_basis"]
        r_max = kw["r_max"]
        PolynomialCutoff_p = kw["PolynomialCutoff_p"]
        feature_irreps_hidden = kw["feature_irreps_hidden"]
        num_layers = kw["num_layers"]
        nonlinearity_type = kw["nonlinearity_type"]
        readout = kw["readout"]
        conv_to_output_hidden_irreps_out = kw["conv_to_output_hidden_irreps_out"]

        super(NeqlipNN, self).__init__()

        self.try_add_edge = True

        # default node_attrs num_types x 0e
        self.one_hot = OneHotAtomEncoding(
            irreps_in=None,
            out_field=AtomicDataDict.NODE_FEATURES_KEY,
            num_types=num_types,
            copy_features=True,
        )

        irreps_temp = self.one_hot.irreps_out

        # default edge_attrs out = irreps_edge_sh=o3.Irreps.spherical_harmonics(3, p=-1),
        self.spharm_edges = SphericalHarmonicEdgeAttrs(irreps_in=irreps_temp,
                                                       out_field=AtomicDataDict.EDGE_ATTRS_KEY,
                                                       irreps_edge_sh=irreps_edge_sh,
                                                       edge_sh_normalization="component",
                                                       edge_sh_normalize=True)

        irreps_temp = self.spharm_edges.irreps_out

        # default edge_embedding num_basis x 0e
        self.radial_basis = RadialBasisEdgeEncoding(irreps_in=irreps_temp,
                                                    out_field=AtomicDataDict.EDGE_EMBEDDING_KEY,
                                                    basis=BesselBasis, cutoff=PolynomialCutoff,
                                                    basis_kwargs={"r_max": r_max, "num_basis": num_basis},
                                                    cutoff_kwargs={"r_max": r_max, "p": PolynomialCutoff_p},
                                                    )

        irreps_temp = self.radial_basis.irreps_out

        self.chemical_embedding = AtomwiseLinear(in_field=AtomicDataDict.NODE_FEATURES_KEY,
                                                 out_field=AtomicDataDict.NODE_FEATURES_KEY,
                                                 irreps_in=irreps_temp,
                                                 irreps_out=feature_irreps_hidden)

        irreps_temp = self.chemical_embedding.irreps_out

        for i in range(num_layers):
            ncl = ConvNetLayer(
                irreps_in=irreps_temp, feature_irreps_hidden=feature_irreps_hidden,
                convolution=InteractionBlock, nonlinearity_type=nonlinearity_type,
                resnet=False,
                convolution_kwargs={
                    "irreps_in": irreps_temp,
                    "irreps_out": irreps_temp,
                    "invariant_layers": 2,
                    "invariant_neurons": 32,
                    "avg_num_neighbors": None,
                    "use_sc": True,
                    "nonlinearity_scalars": {"e": "silu"}, }, )
            setattr(self, f"layer{i}_convnet", ncl)

        irreps_temp = self.chemical_embedding.irreps_out

        self.conv_to_output_hidden = AtomwiseLinear(in_field=AtomicDataDict.NODE_FEATURES_KEY,
                                                    out_field=AtomicDataDict.NODE_FEATURES_KEY,
                                                    irreps_in=irreps_temp,
                                                    irreps_out=conv_to_output_hidden_irreps_out)  # don't change size

        irreps_temp = self.conv_to_output_hidden.irreps_out

        self.output_hidden_to_scalar = AtomwiseLinear(in_field=AtomicDataDict.NODE_FEATURES_KEY,
                                                      out_field=AtomicDataDict.PER_ATOM_ENERGY_KEY,
                                                      irreps_in=irreps_temp, irreps_out="1x0e")  # scalar

        irreps_temp = self.output_hidden_to_scalar.irreps_out
        self.output_reduce = AtomwiseReduce(in_field=AtomicDataDict.PER_ATOM_ENERGY_KEY,
                                            out_field=AtomicDataDict.TOTAL_ENERGY_KEY,
                                            reduce=readout,
                                            irreps_in=None, )


        self._init_irreps(
            irreps_in={},
            irreps_out=irreps_temp.copy(),
        )

    def forward(self, data: Union[Batch, Data, dict]):

        if not isinstance(data, dict):
            data = to_atomicdatadict(data)

        if self.try_add_edge:
            data = AtomicDataDict.with_edge_vectors(data, with_lengths=True)

        for module in self.children():
            data = module(data)

        data[AtomicDataDict.TOTAL_ENERGY_KEY] = data[AtomicDataDict.TOTAL_ENERGY_KEY].view(-1)
        return data



# from typing import List, NamedTuple, Optional, Tuple, Union
#
# from opt_einsum_fx import optimize_einsums_full
# import torch
# from torch import fx
#
# import e3nn
# from e3nn import o3
# from e3nn.util import prod
# from e3nn.util.codegen import CodeGenMixin
# from e3nn.util.jit import compile_mode
#
# from e3nn.o3._tensor_product._codegen import _sum_tensors
#
#
# class Instruction(NamedTuple):
#     i_in: int
#     i_out: int
#     path_shape: tuple
#     path_weight: float
#
#
# @compile_mode("script")
# class Linear(CodeGenMixin, torch.nn.Module):
#     r"""Linear operation equivariant to :math:`O(3)`
#
#     Notes
#     -----
#         `e3nn.o3.Linear` objects created with different partitionings of the same irreps, such as ``Linear("10x0e", "0e")``
#         and ``Linear("3x0e + 7x0e", "0e")``, are *not* equivalent: the second module has more instructions, which affects
#         normalization. In a rough sense:
#
#             Linear("10x0e", "0e") = normalization_coeff_0 * W_0 @ input
#             Linear("3x0e + 7x0e", "0e") = normalization_coeff_1 * W_1 @ input[:3] + normalization_coeff_2 * W_2 @ input[3:]
#
#         To make them equivalent, simplify ``irreps_in`` before constructing network modules:
#
#             o3.Irreps("3x0e + 7x0e").simplify()  # => 10x0e
#
#
#     Parameters
#     ----------
#     irreps_in : `e3nn.o3.Irreps`
#         representation of the input
#
#     irreps_out : `e3nn.o3.Irreps`
#         representation of the output
#
#     internal_weights : bool
#         whether the `e3nn.o3.Linear` should store its own weights. Defaults to ``True`` unless ``shared_weights`` is
#         explicitly set to ``False``, for consistancy with `e3nn.o3.TensorProduct`.
#
#     shared_weights : bool
#         whether the `e3nn.o3.Linear` should be weighted individually for each input in a batch. Defaults to ``False``.
#         Cannot be ``True`` if ``internal_weights`` is ``True``.
#
#     instructions : list of 2-tuples, optional
#         list of tuples ``(i_in, i_out)`` indicating which irreps in ``irreps_in`` should contribute to which irreps in
#         ``irreps_out``. If ``None`` (the default), all allowable instructions will be created: every ``(i_in, i_out)`` such
#         that ``irreps_in[i_in].ir == irreps_out[i_out].ir``.
#
#     biases : list of bool, optional
#         indicates for each element of ``irreps_out`` if it has a bias. By default there is no bias.
#         If ``biases=True`` it gives bias to all scalars (l=0 and p=1).
#
#     Attributes
#     ----------
#     weight_numel : int
#         the size of the weights for this `e3nn.o3.Linear`
#
#     Examples
#     --------
#     Linearly combines 4 scalars into 8 scalars and 16 vectors into 8 vectors.
#
#     >>> lin = Linear("4x0e+16x1o", "8x0e+8x1o")
#     >>> lin.weight_numel
#     160
#
#     Create a "block sparse" linear that does not combine two different groups of scalars;
#     note that the number of weights is 4*4 + 3*3 = 25:
#
#     >>> lin = Linear("4x0e + 3x0e", "4x0e + 3x0e", instructions=[(0, 0), (1, 1)])
#     >>> lin.weight_numel
#     25
#
#     Be careful: because they have different instructions, the following two operations are not normalized in the same way,
#     even though they contain all the same "connections":
#
#     >>> lin1 = Linear("10x0e", "0e")
#     >>> lin2 = Linear("3x0e + 7x0e", "0e")
#     >>> lin1.weight_numel == lin2.weight_numel
#     True
#     >>> with torch.no_grad():
#     ...     lin1.weight.fill_(1.0)
#     ...     lin2.weight.fill_(1.0)
#     Parameter containing:
#     ...
#     >>> x = torch.arange(10.0)
#     >>> (lin1(x) - lin2(x)).abs().item() < 1e-5
#     True
#
#     """
#     weight_numel: int
#     internal_weights: bool
#     shared_weights: bool
#
#     def __init__(
#         self,
#         irreps_in,
#         irreps_out,
#         *,
#         f_in: Optional[int] = None,
#         f_out: Optional[int] = None,
#         internal_weights: Optional[bool] = None,
#         shared_weights: Optional[bool] = None,
#         instructions: Optional[List[Tuple[int, int]]] = None,
#         biases: Union[bool, List[bool]] = False,
#         path_normalization: str = "element",
#         _optimize_einsums: Optional[bool] = None,
#     ):
#         super().__init__()
#
#         assert path_normalization in ["element", "path"]
#
#         irreps_in = o3.Irreps(irreps_in)
#         irreps_out = o3.Irreps(irreps_out)
#
#         if instructions is None:
#             # By default, make all possible connections
#             instructions = [
#                 (i_in, i_out)
#                 for i_in, (_, ir_in) in enumerate(irreps_in)
#                 for i_out, (_, ir_out) in enumerate(irreps_out)
#                 if ir_in == ir_out
#             ]
#
#         instructions = [
#             Instruction(
#                 i_in=i_in,
#                 i_out=i_out,
#                 path_shape=(irreps_in[i_in].mul, irreps_out[i_out].mul),
#                 path_weight=1,
#             )
#             for i_in, i_out in instructions
#         ]
#
#         def alpha(ins):
#             x = sum(
#                 irreps_in[i.i_in if path_normalization == "element" else ins.i_in].mul
#                 for i in instructions
#                 if i.i_out == ins.i_out
#             )
#             if f_in is not None:
#                 x *= f_in
#             return 1.0 if x == 0 else x
#
#         instructions = [
#             Instruction(i_in=ins.i_in, i_out=ins.i_out, path_shape=ins.path_shape, path_weight=alpha(ins) ** (-0.5))
#             for ins in instructions
#         ]
#
#         for ins in instructions:
#             if not ins.i_in < len(irreps_in):
#                 raise IndexError(f"{ins.i_in} is not a valid index for irreps_in")
#             if not ins.i_out < len(irreps_out):
#                 raise IndexError(f"{ins.i_out} is not a valid index for irreps_out")
#             if not (ins.i_in == -1 or irreps_in[ins.i_in].ir == irreps_out[ins.i_out].ir):
#                 raise ValueError(f"{ins.i_in} and {ins.i_out} do not have the same irrep")
#
#         if biases is None:
#             biases = len(irreps_out) * (False,)
#         if isinstance(biases, bool):
#             biases = [biases and ir.is_scalar() for _, ir in irreps_out]
#
#         assert len(biases) == len(irreps_out)
#         assert all(ir.is_scalar() or (not b) for b, (_, ir) in zip(biases, irreps_out))
#
#         instructions += [
#             Instruction(i_in=-1, i_out=i_out, path_shape=(mul_ir.dim,), path_weight=1.0)
#             for i_out, (bias, mul_ir) in enumerate(zip(biases, irreps_out))
#             if bias
#         ]
#
#         # == Process arguments ==
#         if shared_weights is False and internal_weights is None:
#             internal_weights = False
#
#         if shared_weights is None:
#             shared_weights = True
#
#         if internal_weights is None:
#             internal_weights = True
#
#         assert shared_weights or not internal_weights
#         self.internal_weights = internal_weights
#         self.shared_weights = shared_weights
#
#         self.irreps_in = irreps_in
#         self.irreps_out = irreps_out
#         self.instructions = instructions
#
#         opt_defaults = e3nn.get_optimization_defaults()
#         self._optimize_einsums = _optimize_einsums if _optimize_einsums is not None else opt_defaults["optimize_einsums"]
#         del opt_defaults
#
#         # == Generate code ==
#         graphmod, self.weight_numel, self.bias_numel = _codegen_linear(
#             self.irreps_in,
#             self.irreps_out,
#             self.instructions,
#             f_in,
#             f_out,
#             shared_weights=shared_weights,
#             optimize_einsums=self._optimize_einsums,
#         )
#         self._codegen_register({"_compiled_main": graphmod})
#
#         # == Generate weights ==
#         if internal_weights and self.weight_numel > 0:
#             assert self.shared_weights, "Having internal weights impose shared weights"
#             self.weight = torch.nn.Parameter(torch.randn(*((f_in, f_out) if f_in is not None else ()), self.weight_numel))
#         else:
#             # For TorchScript, there always has to be some kind of defined .weight
#             self.register_buffer("weight", torch.Tensor())
#
#         # == Generate biases ==
#         if internal_weights and self.bias_numel > 0:
#             assert self.shared_weights, "Having internal weights impose shared weights"
#             self.bias = torch.nn.Parameter(
#                 torch.zeros(*((f_out,) if f_out is not None else ()), self.bias_numel)
#             )  # see appendix C.1 and Eq.5 of https://arxiv.org/pdf/2011.14522.pdf
#         else:
#             self.register_buffer("bias", torch.Tensor())
#
#         # == Compute output mask ==
#         if self.irreps_out.dim > 0:
#             output_mask = torch.cat(
#                 [
#                     torch.ones(mul_ir.dim)
#                     if any((ins.i_out == i_out) and (0 not in ins.path_shape) for ins in self.instructions)
#                     else torch.zeros(mul_ir.dim)
#                     for i_out, mul_ir in enumerate(self.irreps_out)
#                 ]
#             )
#         else:
#             output_mask = torch.ones(0)
#         self.register_buffer("output_mask", output_mask)
#
#     def __repr__(self):
#         return f"{self.__class__.__name__}({self.irreps_in} -> {self.irreps_out} | {self.weight_numel} weights)"
#
#     def forward(self, features, weight: Optional[torch.Tensor] = None, bias: Optional[torch.Tensor] = None):
#         """evaluate
#
#         Parameters
#         ----------
#         features : `torch.Tensor`
#             tensor of shape ``(..., irreps_in.dim)``
#
#         weight : `torch.Tensor`, optional
#             required if ``internal_weights`` is `False`
#
#         Returns
#         -------
#         `torch.Tensor`
#             tensor of shape ``(..., irreps_out.dim)``
#         """
#         if weight is None:
#             if self.weight_numel > 0 and not self.internal_weights:
#                 raise RuntimeError("Weights must be provided when internal_weights = False")
#             weight = self.weight
#         if bias is None:
#             if self.bias_numel > 0 and not self.internal_weights:
#                 raise RuntimeError("Biases must be provided when internal_weights = False")
#             bias = self.bias
#         return self._compiled_main(features, weight, bias)
#
#     def weight_view_for_instruction(self, instruction: int, weight: Optional[torch.Tensor] = None) -> torch.Tensor:
#         r"""View of weights corresponding to ``instruction``.
#
#         Parameters
#         ----------
#         instruction : int
#             The index of the instruction to get a view on the weights for.
#
#         weight : `torch.Tensor`, optional
#             like ``weight`` argument to ``forward()``
#
#         Returns
#         -------
#         `torch.Tensor`
#             A view on ``weight`` or this object's internal weights for the weights corresponding to the ``instruction`` th
#             instruction.
#         """
#         if weight is None:
#             assert self.internal_weights, "Weights must be provided when internal_weights = False"
#             weight = self.weight
#         batchshape = weight.shape[:-1]
#         offset = sum(prod(ins.path_shape) for ins in self.instructions[:instruction])
#         ins = self.instructions[instruction]
#         return weight.narrow(-1, offset, prod(ins.path_shape)).view(batchshape + ins.path_shape)
#
#     def weight_views(self, weight: Optional[torch.Tensor] = None, yield_instruction: bool = False):
#         r"""Iterator over weight views for all instructions.
#
#         Parameters
#         ----------
#         weight : `torch.Tensor`, optional
#             like ``weight`` argument to ``forward()``
#
#         yield_instruction : `bool`, default False
#             Whether to also yield the corresponding instruction.
#
#         Yields
#         ------
#         If ``yield_instruction`` is ``True``, yields ``(instruction_index, instruction, weight_view)``.
#         Otherwise, yields ``weight_view``.
#         """
#         if weight is None:
#             assert self.internal_weights, "Weights must be provided when internal_weights = False"
#             weight = self.weight
#         batchshape = weight.shape[:-1]
#         offset = 0
#         for ins_i, ins in enumerate(self.instructions):
#             flatsize = prod(ins.path_shape)
#             this_weight = weight.narrow(-1, offset, flatsize).view(batchshape + ins.path_shape)
#             offset += flatsize
#             if yield_instruction:
#                 yield ins_i, ins, this_weight
#             else:
#                 yield this_weight
#
#
# def _codegen_linear(
#     irreps_in: o3.Irreps,
#     irreps_out: o3.Irreps,
#     instructions: List[Instruction],
#     f_in: Optional[int] = None,
#     f_out: Optional[int] = None,
#     shared_weights: bool = False,
#     optimize_einsums: bool = True,
# ) -> Tuple[fx.GraphModule, int, int]:
#     graph_out = fx.Graph()
#     tracer_out = fx.proxy.GraphAppendingTracer(graph_out)
#
#     # = Function definitions =
#     x = fx.Proxy(graph_out.placeholder("x", torch.Tensor), tracer_out)
#     ws = fx.Proxy(graph_out.placeholder("w", torch.Tensor), tracer_out)
#     bs = fx.Proxy(graph_out.placeholder("b", torch.Tensor), tracer_out)
#
#     if f_in is None:
#         size = x.shape[:-1]
#         outsize = size + (irreps_out.dim,)
#     else:
#         size = x.shape[:-2]
#         outsize = size + (
#             f_out,
#             irreps_out.dim,
#         )
#
#     bias_numel = sum(irreps_out[i.i_out].dim for i in instructions if i.i_in == -1)
#     if bias_numel > 0:
#         if f_out is None:
#             bs = bs.reshape(-1, bias_numel)
#         else:
#             bs = bs.reshape(-1, f_out, bias_numel)
#
#     # = Short-circut for nothing to do =
#     # We produce no code for empty instructions
#     instructions = [ins for ins in instructions if 0 not in ins.path_shape]
#
#     if len(instructions) == 0 and bias_numel == 0:
#         out = x.new_zeros(outsize)
#
#         graph_out.output(out.node, torch.Tensor)
#         # Short circut
#         # 0 is weight_numel
#         mx = fx.GraphModule({}, graph_out, "linear_forward")
#         # fixed bug for fx.Graph when torch-script bug
#         mx.meta.update({"fixed_bug":"temp"})
#         return mx, 0, 0
#
#     if f_in is None:
#         x = x.reshape(-1, irreps_in.dim)
#     else:
#         x = x.reshape(-1, f_in, irreps_in.dim)
#     batch_out = x.shape[0]
#
#     weight_numel = sum(prod(ins.path_shape) for ins in instructions if ins.i_in != -1)
#     if weight_numel > 0:
#         ws = ws.reshape(-1, weight_numel) if f_in is None else ws.reshape(-1, f_in, f_out, weight_numel)
#
#     # = extract individual input irreps =
#     if len(irreps_in) == 1:
#         x_list = [x.reshape(batch_out, *(() if f_in is None else (f_in,)), irreps_in[0].mul, irreps_in[0].ir.dim)]
#     else:
#         x_list = [
#             x.narrow(-1, i.start, mul_ir.dim).reshape(batch_out, *(() if f_in is None else (f_in,)), mul_ir.mul, mul_ir.ir.dim)
#             for i, mul_ir in zip(irreps_in.slices(), irreps_in)
#         ]
#
#     z = "" if shared_weights else "z"
#
#     flat_weight_index = 0
#     flat_bias_index = 0
#
#     out_list = []
#
#     for ins in instructions:
#         mul_ir_out = irreps_out[ins.i_out]
#
#         if ins.i_in == -1:
#             # = bias =
#             b = bs.narrow(-1, flat_bias_index, prod(ins.path_shape))
#             flat_bias_index += prod(ins.path_shape)
#             out_list += [(ins.path_weight * b).reshape(1, *(() if f_out is None else (f_out,)), mul_ir_out.dim)]
#         else:
#             mul_ir_in = irreps_in[ins.i_in]
#
#             # Short-circut for empty irreps
#             if mul_ir_in.dim == 0 or mul_ir_out.dim == 0:
#                 continue
#
#             # Extract the weight from the flattened weight tensor
#             path_nweight = prod(ins.path_shape)
#             if len(instructions) == 1:
#                 # Avoid unnecessary view when there is only one weight
#                 w = ws
#             else:
#                 w = ws.narrow(-1, flat_weight_index, path_nweight)
#             w = w.reshape((() if shared_weights else (-1,)) + (() if f_in is None else (f_in, f_out)) + ins.path_shape)
#             flat_weight_index += path_nweight
#
#             if f_in is None:
#                 ein_out = torch.einsum(f"{z}uw,zui->zwi", w, x_list[ins.i_in])
#             else:
#                 ein_out = torch.einsum(f"{z}xyuw,zxui->zywi", w, x_list[ins.i_in])
#
#             ein_out = ins.path_weight * ein_out
#
#             out_list += [ein_out.reshape(batch_out, *(() if f_out is None else (f_out,)), mul_ir_out.dim)]
#
#     # = Return the result =
#     out = [
#         _sum_tensors(
#             [out for ins, out in zip(instructions, out_list) if ins.i_out == i_out],
#             shape=(batch_out, *(() if f_out is None else (f_out,)), mul_ir_out.dim),
#             like=x,
#         )
#         for i_out, mul_ir_out in enumerate(irreps_out)
#         if mul_ir_out.mul > 0
#     ]
#     if len(out) > 1:
#         out = torch.cat(out, dim=-1)
#     else:
#         out = out[0]
#
#     out = out.reshape(outsize)
#
#     graph_out.output(out.node, torch.Tensor)
#
#     # check graphs
#     graph_out.lint()
#
#     graphmod_out = fx.GraphModule({}, graph_out, "linear_forward")
#     # fixed bug for fx.Graph when torch-script bug
#     graphmod_out.meta.update({"fixed_bug": "temp"})
#
#     # TODO: when eliminate_dead_code() is in PyTorch stable, use that
#     if optimize_einsums:
#         # See _tensor_product/_codegen.py for notes
#         batchdim = 4
#         example_inputs = (
#             torch.zeros((batchdim, *(() if f_in is None else (f_in,)), irreps_in.dim)),
#             torch.zeros(
#                 1 if shared_weights else batchdim,
#                 f_in or 1,
#                 f_out or 1,
#                 weight_numel,
#             ),
#             torch.zeros(
#                 1 if shared_weights else batchdim,
#                 f_out or 1,
#                 bias_numel,
#             ),
#         )
#         graphmod_out = optimize_einsums_full(graphmod_out, example_inputs)
#
#
#
#     return graphmod_out, weight_numel, bias_numel
#
#
# if __name__=="__main__":
#     lin1 = Linear("10x0e", "0e")
#     lin2 = Linear("3x0e + 7x0e", "0e")
#     lin1.weight_numel == lin2.weight_numel



""" Interaction Block """
from typing import Optional

import torch
from e3nn import o3
from e3nn.nn import FullyConnectedNet
from e3nn.o3 import TensorProduct, Linear, FullyConnectedTensorProduct
from torch_runstats.scatter import scatter

from hpc2ml.data.nequlip_support import AtomicDataDict
from hpc2ml.data.nequlip_support.AtomicDataDict import ST
from ._graph_mixin import GraphModuleMixin
from .nonlinearities import ShiftedSoftPlus


class InteractionBlock(GraphModuleMixin, torch.nn.Module):
    avg_num_neighbors: Optional[float]
    use_sc: bool

    def __init__(
            self,
            irreps_in,
            irreps_out,
            invariant_layers=1,
            invariant_neurons=8,
            avg_num_neighbors=None,
            use_sc=True,
            nonlinearity_scalars=None,
    ) -> None:
        """
        InteractionBlock.

        :param irreps_node_attr: Nodes attribute irreps
        :param irreps_edge_attr: Edge attribute irreps
        :param irreps_out: Output irreps, in our case typically a single scalar
        :param radial_layers: Number of radial layers, default = 1
        :param radial_neurons: Number of hidden neurons in radial function, default = 8
        :param avg_num_neighbors: Number of neighbors to divide by, default None => no normalization.
        :param number_of_basis: Number or Basis function, default = 8
        :param irreps_in: Input Features, default = None
        :param use_sc: bool, use self-connection or not
        """
        super().__init__()

        if nonlinearity_scalars is None:
            nonlinearity_scalars = {"e": "silu"}
        self._init_irreps(
            irreps_in=irreps_in,
            required_irreps_in=[
                AtomicDataDict.EDGE_EMBEDDING_KEY,
                AtomicDataDict.EDGE_ATTRS_KEY,
                AtomicDataDict.NODE_FEATURES_KEY,
                AtomicDataDict.NODE_ATTRS_KEY,
            ],
            my_irreps_in={
                AtomicDataDict.EDGE_EMBEDDING_KEY: o3.Irreps(
                    [
                        (
                            irreps_in[AtomicDataDict.EDGE_EMBEDDING_KEY].num_irreps,
                            (0, 1),
                        )
                    ]  # (0, 1) is even (invariant) scalars.
                    # We are forcing the EDGE_EMBEDDING to be invariant scalars so we can use a dense network
                )
            },
            irreps_out={AtomicDataDict.NODE_FEATURES_KEY: irreps_out},
        )

        self.avg_num_neighbors = avg_num_neighbors
        self.use_sc = use_sc

        feature_irreps_in = self.irreps_in[AtomicDataDict.NODE_FEATURES_KEY]
        feature_irreps_out = self.irreps_out[AtomicDataDict.NODE_FEATURES_KEY]
        irreps_edge_attr = self.irreps_in[AtomicDataDict.EDGE_ATTRS_KEY]

        # - Build modules -
        self.linear_1 = Linear(
            irreps_in=feature_irreps_in,
            irreps_out=feature_irreps_in,
            internal_weights=True,
            shared_weights=True,
        )

        irreps_mid = []
        instructions = []

        for i, (mul, ir_in) in enumerate(feature_irreps_in):
            for j, (_, ir_edge) in enumerate(irreps_edge_attr):
                for ir_out in ir_in * ir_edge:
                    if ir_out in feature_irreps_out:
                        k = len(irreps_mid)
                        irreps_mid.append((mul, ir_out))
                        instructions.append((i, j, k, "uvu", True))

        # We sort the output irreps of the tensor product so that we can simplify them
        # when they are provided to the second o3.Linear
        irreps_mid = o3.Irreps(irreps_mid)
        irreps_mid, p, _ = irreps_mid.sort()

        # Permute the output indexes of the instructions to match the sorted irreps:
        instructions = [
            (i_in1, i_in2, p[i_out], mode, train)
            for i_in1, i_in2, i_out, mode, train in instructions
        ]

        tp = TensorProduct(
            feature_irreps_in,
            irreps_edge_attr,
            irreps_mid,
            instructions,
            shared_weights=False,
            internal_weights=False,
        )

        # init_irreps already confirmed that the edge embeddding is all invariant scalars
        self.fc = FullyConnectedNet(
            [self.irreps_in[AtomicDataDict.EDGE_EMBEDDING_KEY].num_irreps]
            + invariant_layers * [invariant_neurons]
            + [tp.weight_numel],
            {
                "ssp": ShiftedSoftPlus,
                "silu": torch.nn.functional.silu,
            }[nonlinearity_scalars["e"]],
        )

        self.tp = tp

        self.linear_2 = Linear(
            # irreps_mid has uncoallesed irreps because of the uvu instructions,
            # but there's no reason to treat them seperately for the Linear
            # Note that normalization of o3.Linear changes if irreps are coallesed
            # (likely for the better)
            irreps_in=irreps_mid.simplify(),
            irreps_out=feature_irreps_out,
            internal_weights=True,
            shared_weights=True,
        )

        self.sc = None
        if self.use_sc:
            self.sc = FullyConnectedTensorProduct(
                feature_irreps_in,
                self.irreps_in[AtomicDataDict.NODE_ATTRS_KEY],
                feature_irreps_out,
            )

    def forward(self, data: ST) -> ST:
        """
        Evaluate interaction Block with ResNet (self-connection).

        :param node_input:
        :param node_attr:
        :param edge_src:
        :param edge_dst:
        :param edge_attr:
        :param edge_length_embedded:

        :return:
        """
        weight = self.fc(data[AtomicDataDict.EDGE_EMBEDDING_KEY])

        x = data[AtomicDataDict.NODE_FEATURES_KEY]
        edge_src = data[AtomicDataDict.EDGE_INDEX_KEY][1]
        edge_dst = data[AtomicDataDict.EDGE_INDEX_KEY][0]

        if self.sc is not None:
            sc = self.sc(x, data[AtomicDataDict.NODE_ATTRS_KEY])

        x = self.linear_1(x)
        edge_features = self.tp(
            x[edge_src], data[AtomicDataDict.EDGE_ATTRS_KEY], weight
        )
        x = scatter(edge_features, edge_dst, dim=0, dim_size=len(x))

        # Necessary to get TorchScript to be able to type infer when its not None
        avg_num_neigh: Optional[float] = self.avg_num_neighbors
        if avg_num_neigh is not None:
            x = x.div(avg_num_neigh ** 0.5)

        x = self.linear_2(x)

        if self.sc is not None:
            x = x + sc

        data[AtomicDataDict.NODE_FEATURES_KEY] = x
        return data



from functools import lru_cache
from math import cos, sin

import scipy

from scipy.ndimage import affine_transform
import numpy as np


@lru_cache(maxsize=10)
def get_matrix(angles=(90, 90, 90), inverse=False):
    """
    Axis of rotation Get matrix by angle.
    (shear+compress)
    Examples: z = 120
    
    ############################################################
    
    ----------------------          --------------------------------
    -oooooooooooooooooooo-          --------------------------------
    -oooooooooooooooooooo-          -oooooooooooooooooooo-----------
    -oooooooooooooooooooo-          ---oooooooooooooooooooo---------
    -oooooooooooooooooooo-    >>>   -----oooooooooooooooooooo-------
    -oooooooooooooooooooo-          -------oooooooooooooooooooo-----
    -oooooooooooooooooooo-          ---------oooooooooooooooooooo---
    -oooooooooooooooooooo-          -----------oooooooooooooooooooo-
    ----------------------          --------------------------------
    
    ############################################################
    
    1.The ``matrix`` is the transform matrix to rotate the data with angle. Always in Cartesian coordinates.

    2.The ``inverse matrix`` is the interpolation matrix for get true data matrix(Cartesian coordinates)
     from relative data matrix (Non-Cartesian coordinates).
     The

    Parameters
    ----------
    angles: tuple
        3 angle of x, y, z
        z angle is the intersection angle of x,y,
        y angle is the intersection angle of x,z,
        x angle is the intersection angle of y,z.
    inverse:
        Compute the (multiplicative) inverse of a matrix.

    """
    theta1, theta2, theta3 = [np.pi / 180 * angle for angle in angles]

    matrix1 = np.array([[1, cos(theta3), 0],
                        [0, sin(theta3), 0],
                        [0, 0, 1]])

    matrix2 = np.array([[1, 0, 0],
                        [0, 1, cos(theta1)],
                        [0, 0, sin(theta1)]])

    matrix3 = np.array([[1, 0, cos(theta2)],
                        [0, 1, 0],
                        [0, 0, sin(theta2)]])
    matrix = np.dot(matrix1, matrix2).dot(matrix3)

    if inverse:
        matrix = np.linalg.inv(matrix)

    return matrix


def rotation_axis_by_angle(data, angles=(90, 90, 90), times=(2, 2, 2)):
    """
    Get true data matrix(Cartesian coordinates) from relative data matrix (Non-Cartesian coordinates).

    Parameters
    ----------
    data: np.ndarray
        data with shape (nx,ny,nz).
    angles:tuple
        3 angle of x, y, z
        z angle is the intersection angle of x,y,
        y angle is the intersection angle of x,z,
        x angle is the intersection angle of y,z.
    times: tuple
        expand the multiple of the matrix.

    """
    matrix = get_matrix(angles=angles, inverse=True)
    return rotation_axis_by_matrix(data, matrix, times=times)


def rotation_axis_by_matrix(data, matrix, times=(2, 2, 2)):
    """
    Get true data matrix(Cartesian coordinates) from relative data matrix (Non-Cartesian coordinates).

    Parameters
    ----------
    data: np.ndarray
        data with shape (nx,ny,nz).
    matrix:tuple
        See Also ``get_matrix``
    times: tuple
        expand the multiple of the matrix.

    """
    dims_old = data.shape

    dims = tuple([int(i * j) for i, j in zip(dims_old, times)])
    n_data = np.zeros(dims)
    d0s = int((dims[0] - dims_old[0]) / 2)
    d1s = int((dims[1] - dims_old[1]) / 2)
    d2s = int((dims[2] - dims_old[2]) / 2)
    n_data[d0s:d0s + dims_old[0], d1s:d1s + dims_old[1], d2s:d2s + dims_old[2]] = data
    coords = np.meshgrid(range(dims[0]), range(dims[1]), range(dims[2]), indexing="ij")
    xy_coords = np.vstack([coords[0].reshape(-1), coords[1].reshape(-1), coords[2].reshape(-1)])
    # apply the transformation matrix
    # please note: the coordinates are not homogeneous.
    # for the 3D case, I've added code for homogeneous coordinates, you might want to look at that
    # please also note: rotation is always around the origin:
    # since I want the origin to be in the image center, I had to substract dim/2, rotate, then add it again
    dims2 = np.array([i / 2 for i in dims])
    dims2 = dims2.reshape(-1, 1)
    xy_coords = np.dot(matrix, xy_coords - dims2) + dims2
    #
    # # undo the stacking and reshaping
    x = xy_coords[0, :]
    y = xy_coords[1, :]
    z = xy_coords[2, :]
    x = x.reshape(dims, order="A")
    y = y.reshape(dims, order="A")
    z = z.reshape(dims, order="A")

    new_coords = [x, y, z]

    # use map_coordinates to sample values for the new image
    new_img = scipy.ndimage.map_coordinates(n_data, new_coords, order=2)
    return new_img


def _coords(points, angles=(90, 90, 90), times=(2, 2, 2)):
    """

    Parameters
    ----------
    points: np.darray
        percent of shape.
        key points with shape(n_sample,3)

    angles:tuple
        3 angle of x, y, z
        z angle is the intersection angle of x,y,
        y angle is the intersection angle of x,z,
        x angle is the intersection angle of y,z.
    times: tuple
        expand the multiple of the matrix.

    """
    dims_old = [1, 1, 1]
    matrix = get_matrix(angles=angles)

    times = np.array(list(times))
    times = times.reshape((-1, 1))
    dims_old = np.array(dims_old)
    dims_old = dims_old.reshape(-1, 1)
    dims2 = dims_old / 2

    points = points.T * dims_old

    xy_coords = np.dot(matrix, points - dims2) + dims2

    xy_coords = xy_coords + (times / 2 - 0.5)

    return xy_coords


def rote_index(points, data, angles=(90, 90, 90), times=(2, 2, 2), data_init=True, return_type="float"):
    """

    Parameters
    ----------
    points: np.darray
        key points with shape(n_sample,3)
        percent of shape.
    data: np.ndarray or tuple
        data or data.shape
    data_init:bool
        The data is the init data (relative location) or Cartesian coordinates.(rotation_axis_by_angle)
    angles:tuple
        3 angle of x, y, z
        z angle is the intersection angle of x,y,
        y angle is the intersection angle of x,z,
        x angle is the intersection angle of y,z.
    times: tuple
        expand the multiple of the matrix.
    return_type:str
        "float", "int", "percent"
        for "float", "int" return the new index
        for "percent" return the new percent.

    """

    data_shape = data.shape if isinstance(data, np.ndarray) else data
    if data_init:
        times_np = np.array([1,1,1])
    else:
        times_np = np.array(times)

    dims = data_shape
    dims = np.array(dims).reshape((-1, 1))

    xy_coords = _coords(points, angles=angles, times=times)
    if return_type == "percent":
        return xy_coords
    if return_type == "float":
        return (dims * xy_coords/times_np).T
    else:
        return np.round((dims * xy_coords/times_np).T).astype(int) # for rounding off: .4 -, .5 +


def rote_value(points, data, angles=(90, 90, 90), times=(2, 2, 2), method="in", data_type="td"):
    """

    Parameters
    ----------
    points: np.darray
        key points with shape(n_sample,3)
        percent of shape.
    data: np.ndarray
        data
    angles:tuple
        3 angle of x, y, z
        z angle is the intersection angle of x,y,
        y angle is the intersection angle of x,z,
        x angle is the intersection angle of y,z.
    times: tuple
        expand the multiple of the matrix.
    data_type:str
        if  "init"  the data accept init data (elfcar, chgcar). see rotation_axis_by_angle.
        if  "td"  the data accept true matrix data . see rotation_axis_by_angle.

    method:str
        if  "near" , return nearest site's value.
        if  "inter" , return the interpolation value.
    """

    if data_type == "td":
        new_data = data
    else:
        new_data = rotation_axis_by_angle(data, angles=angles, times=times)

    if method == "near":
        ind = rote_index(points, data, angles=angles, times=times, return_type="int")
        new_value = np.array([new_data[tuple(i)] for i in ind.T])
        return new_value
    else:
        ind = rote_index(points, data, angles=angles, times=times, return_type="float")
        new_value = scipy.ndimage.map_coordinates(new_data, ind, order=2)
        return new_value



import warnings
from collections import deque

import numpy as np
from ase.calculators.calculator import PropertyNotImplementedError
from ase.optimize.optimize import Optimizer

from hpc2ml.optimize.basegpmin import BaseGPMin


class EIMin(BaseGPMin):
    """EffectiveImprovementMin"""

    def __init__(self, atoms, restart=None, logfile='-', trajectory=None,
                 master=None, force_consistent=None, converged_std="forces",
                 update_hyperparams=False, batch_size=2):

        self.function_calls = 1
        self.force_calls = 0
        self.x_list = []  # Training set features
        self.y_list = []  # Training set targets

        Optimizer.__init__(self, atoms, restart, logfile,
                           trajectory, master, force_consistent)

        if batch_size is None:
            self.nbatch = 1
        else:
            self.nbatch = batch_size

        self.converged_std = converged_std
        self.update_hp = update_hyperparams
        self.function_calls = 1
        self.force_calls = 0

    def update(self, r, e, f):
        """Update the PES

        Update the training set, the prior and the hyperparameters.
        Finally, train the model
        """
        # update the training set
        self.x_list.append(r)
        f = f.reshape(-1)
        y = np.append(np.array(e).reshape(-1), -f)
        self.y_list.append(y)
        if len(self.y_list) > 32:
            self.x_list = self.x_list[-32:]
            self.y_list = self.y_list[-32:]

    def update_batch(self, rs, es, fs):
        """Update the PES

        Update the training set, the prior and the hyperparameters.
        Finally, train the model
        """
        # update the training set
        [self.x_list.append(ri) for ri in rs]
        y = np.concatenate((np.array(es).reshape(-1, 1), -fs), axis=1)
        [self.y_list.append(yi) for yi in y]

        if len(self.y_list) > 32:
            self.x_list = self.x_list[-32:]
            self.y_list = self.y_list[-32:]

        if (self.update_hp and self.function_calls % self.nbatch == 0 and
                self.function_calls != 0):
            self.train()

    def train(self):
        """This train is on model."""
        if not hasattr(self.atoms.calc, "train"):
            return None
        atom_temp = self.atoms.copy()
        atom_temp.calc = self.atoms.calc

        atomss = []
        energys = []
        forcess = []
        for x, y in zip(self.x_list, self.y_list):
            pos = x
            atom_temp.set_positions(pos.reshape(-1, 3))
            energy = y[0]
            forces = y[1:].reshape(-1, 3)
            atomss.append(atom_temp)
            energys.append(energy)
            forcess.append(forces)

        self.atoms.calc.train(atomss, energy=energys, forces=forcess)

    def predict_mean_std(self, x):
        """In general, this part should use the values from vasp.
        But for simple, we just re-back the values from model."""
        atom_temp = self.atoms.copy()
        atom_temp.set_positions(x.reshape(-1, 3))
        calc = self.atoms.calc

        e = calc.get_property("energy", atoms=atom_temp)
        f = calc.get_property("forces", atoms=atom_temp)

        try:
            estd = calc.get_property("energy_std", atoms=atom_temp)
        except NotImplementedError:
            estd = 0.0
        try:
            fstd = calc.get_property("forces_std", atoms=atom_temp)
        except NotImplementedError:
            fstd = [0.0]*len(self.atoms)*3

        es = np.append(np.array(e).ravel(), np.array(f).ravel(), axis=0)
        fs = np.append(np.array(estd).ravel(), np.array(fstd).ravel(), axis=0)
        e_mean_f_mean_e_std_f_std = np.vstack((es, fs))
        return e_mean_f_mean_e_std_f_std

    def predict_mean_std_all(self, rs):
        """e_mean_f_mean_e_std_f_std.
        shape (n_sample, 2 [mean,std], n_atom+1 [e_ms,f_ms] )"""
        calc = self.atoms.calc
        if hasattr(calc, "calculate_batch"):
            atoms_list = []
            for x in rs:
                atom_temp = self.atoms.copy()
                atom_temp.set_positions(x.reshape(-1, 3))
                atoms_list.append(atom_temp)
            res = calc.calculate_batch(atoms_list)

            e = res["energy"]
            f = res["forces"]

            try:
                estd = res["energy_std"]
            except KeyError:
                estd = np.zeros_like(e)
            try:
                fstd = res["forces_std"]
            except KeyError:
                fstd = np.zeros_like(f)

            es = np.concatenate((np.array(e).reshape(-1, 1), np.array(f)), axis=1)
            fs = np.concatenate((np.array(estd).reshape(-1, 1), np.array(fstd)), axis=1)
            res = np.array([es, fs])
            res = res.transpose((1, 0, 2))
            return res

        else:
            res = [self.predict_mean_std(x) for x in rs]
            return np.array(res)

    def step(self, f=None):

        atoms = self.atoms
        if f is None:
            f = atoms.get_forces()

        fc = self.force_consistent
        r0 = atoms.get_positions().reshape(-1)
        e0 = atoms.get_potential_energy(force_consistent=fc)
        self.update(r0, e0, f)

        r_add = None
        e_add = None
        f_add = None

        count = 0

        self.es = deque([e0], maxlen=3)

        while len(self.es) < 3 or np.std(self.es) > 0.01:  # 0.01 eV energy error

            r_, e_, f_ = self.resample_r_e_f(20,
                                             random_factor=0.01,  # this is fixed, 0.01 is best when test.
                                             )  # train

            if r_add is None:
                pass
            else:
                # Active learning iteration (add the sample with good performance).
                r_ = np.concatenate((r_, r_add), axis=0)
                e_ = np.concatenate((e_, e_add), axis=0)
                f_ = np.concatenate((f_, f_add), axis=0)

            self.update_batch(r_, e_, f_)

            random_factor = max(0.05 * 1/(self.nsteps+1), 0.005)  # decrease the degree of position change with step.

            r_ = self.resample_r(50, random_factor=random_factor)

            res = self.predict_mean_std_all(r_)

            ei = self.ei_balance(res)

            # ei = self.ei_balance(res, loss_coef=1-0.02*self.nsteps)
            # decrease the forces weight with step.
            # this should be just used for forces with
            index_ = np.argsort(-ei)[:10]
            r_add = r_[index_]

            r_add, e_add, f_add = self.r_to_e_f(r_add)

            index_min = np.argmin(e_add)  # best
            e1 = e_add[index_min]
            r1 = r_add[index_min]
            f1 = f_add[index_min]

            # if e1 < self.es[-1]+0.00001:
            if e1 < self.es[-1]:
                self.atoms.set_positions(r1.reshape(-1, 3))
                self.es.append(float(e1))

            self.function_calls += 1
            self.force_calls += 1

            if self.converged(f1.reshape(-1, 3)):
                break

            count += 1
            if count == 20:
                warnings.warn("A descent model could not be built")
                break

            # print(count, self.es[-1], e1)


        self.dump()






import warnings

import numpy as np
from ase.io.jsonio import write_json
from ase.parallel import world
from scipy.optimize import minimize

from hpc2ml.optimize.basegpmin import BaseGPMin


class ResampleGPMin(BaseGPMin):
    def __init__(self, atoms, **kwargs):
        super(ResampleGPMin, self).__init__(atoms, **kwargs)

    def step(self, f=None):
        atoms = self.atoms
        if f is None:
            f = atoms.get_forces()

        fc = self.force_consistent
        r0 = self.atoms.get_positions().reshape(-1)
        e0 = self.atoms.get_potential_energy(force_consistent=fc)
        self.update(r0, e0, f)

        r1 = self.relax_model(r0)
        self.atoms.set_positions(r1.reshape(-1, 3))
        e1 = self.atoms.get_potential_energy(force_consistent=fc)
        f1 = self.atoms.get_forces()
        self.function_calls += 1
        self.force_calls += 1
        count = 0

        while e1 >= e0:
            self.update(r1, e1, f1)

            r_, e_, f_ = self.resample_r_e_f(20, random_factor=0.005)  # predict by calc

            self.update_batch(r_, e_, f_)  # train gp model

            r1 = self.relax_model(r0)
            self.atoms.set_positions(r1.reshape(-1, 3))
            e1 = self.atoms.get_potential_energy(force_consistent=self.force_consistent)
            f1 = self.atoms.get_forces()

            self.function_calls += 1
            self.force_calls += 1

            if self.converged(f1):
                break

            count += 1
            if count == 10:
                warnings.warn("A descent model could not be built")
                break

        self.dump()











    


from typing import Tuple, Union, Sequence, Set

from hpc2ml.data.nequlip_support import AtomicDataDict

PBC = Union[bool, Tuple[bool, bool, bool]]

_DEFAULT_LONG_FIELDS: Set[str] = {
    AtomicDataDict.EDGE_INDEX_KEY,
    AtomicDataDict.ATOMIC_NUMBERS_KEY,
    AtomicDataDict.ATOM_TYPE_KEY,
    AtomicDataDict.BATCH_KEY,
}
_DEFAULT_NODE_FIELDS: Set[str] = {
    AtomicDataDict.POSITIONS_KEY,
    AtomicDataDict.NODE_FEATURES_KEY,
    AtomicDataDict.NODE_ATTRS_KEY,
    AtomicDataDict.ATOMIC_NUMBERS_KEY,
    AtomicDataDict.ATOM_TYPE_KEY,
    AtomicDataDict.FORCE_KEY,
    AtomicDataDict.PER_ATOM_ENERGY_KEY,
    AtomicDataDict.BATCH_KEY,
}
_DEFAULT_EDGE_FIELDS: Set[str] = {
    AtomicDataDict.EDGE_CELL_SHIFT_KEY,
    AtomicDataDict.EDGE_VECTORS_KEY,
    AtomicDataDict.EDGE_LENGTH_KEY,
    AtomicDataDict.EDGE_ATTRS_KEY,
    AtomicDataDict.EDGE_EMBEDDING_KEY,
    AtomicDataDict.EDGE_FEATURES_KEY,
}
_DEFAULT_GRAPH_FIELDS: Set[str] = {
    AtomicDataDict.TOTAL_ENERGY_KEY,
    AtomicDataDict.STRESS_KEY,
    AtomicDataDict.VIRIAL_KEY,
    AtomicDataDict.PBC_KEY,
    AtomicDataDict.CELL_KEY,
}
_NODE_FIELDS: Set[str] = set(_DEFAULT_NODE_FIELDS)
_EDGE_FIELDS: Set[str] = set(_DEFAULT_EDGE_FIELDS)
_GRAPH_FIELDS: Set[str] = set(_DEFAULT_GRAPH_FIELDS)
_LONG_FIELDS: Set[str] = set(_DEFAULT_LONG_FIELDS)


def register_fields(
        node_fields: Sequence[str] = [],
        edge_fields: Sequence[str] = [],
        graph_fields: Sequence[str] = [],
        long_fields: Sequence[str] = [],
) -> None:
    r"""Register fields as being per-atom, per-edge, or per-frame.

    Args:
        node_permute_fields: fields that are equivariant to node permutations.
        edge_permute_fields: fields that are equivariant to edge permutations.
    """
    node_fields: set = set(node_fields)
    edge_fields: set = set(edge_fields)
    graph_fields: set = set(graph_fields)
    allfields = node_fields.union(edge_fields, graph_fields)
    assert len(allfields) == len(node_fields) + len(edge_fields) + len(graph_fields)
    _NODE_FIELDS.update(node_fields)
    _EDGE_FIELDS.update(edge_fields)
    _GRAPH_FIELDS.update(graph_fields)
    _LONG_FIELDS.update(long_fields)
    if len(set.union(_NODE_FIELDS, _EDGE_FIELDS, _GRAPH_FIELDS)) < (
            len(_NODE_FIELDS) + len(_EDGE_FIELDS) + len(_GRAPH_FIELDS)
    ):
        raise ValueError(
            "At least one key was registered as more than one of node, edge, or graph!"
        )


def deregister_fields(*fields: Sequence[str]) -> None:
    r"""Deregister a field registered with ``register_fields``.

    Silently ignores fields that were never registered to begin with.

    Args:
        *fields: fields to deregister.
    """
    for f in fields:
        assert f not in _DEFAULT_NODE_FIELDS, "Cannot deregister built-in field"
        assert f not in _DEFAULT_EDGE_FIELDS, "Cannot deregister built-in field"
        assert f not in _DEFAULT_GRAPH_FIELDS, "Cannot deregister built-in field"
        _NODE_FIELDS.discard(f)
        _EDGE_FIELDS.discard(f)
        _GRAPH_FIELDS.discard(f)



# from hpc2ml.data import AtomicDataDict
# from ..nn import GraphModuleMixin, GradientOutput
# from ..nn import PartialForceOutput as PartialForceOutputModule
# from ..nn import StressOutput as StressOutputModule
#
#
# def ForceOutput(model: GraphModuleMixin) -> GradientOutput:
#     r"""Add forces to a model that predicts energy.
#
#     Args:
#         model: the energy model to wrap. Must have ``AtomicDataDict.TOTAL_ENERGY_KEY`` as an output.
#
#     Returns:
#         A ``GradientOutput`` wrapping ``model``.
#     """
#     if AtomicDataDict.FORCE_KEY in model.irreps_out:
#         raise ValueError("This model already has forces outputs.")
#     return GradientOutput(
#         func=model,
#         of=AtomicDataDict.TOTAL_ENERGY_KEY,
#         wrt=AtomicDataDict.POSITIONS_KEY,
#         out_field=[AtomicDataDict.FORCE_KEY],
#         sign=-1,  # forces is the negative gradient
#     )
#
#
# def PartialForceOutput(model: GraphModuleMixin) -> PartialForceOutputModule:
#     r"""Add forces and partial forces to a model that predicts energy.
#
#     Args:
#         model: the energy model to wrap. Must have ``AtomicDataDict.TOTAL_ENERGY_KEY`` as an output.
#
#     Returns:
#         A ``GradientOutput`` wrapping ``model``.
#     """
#     if (
#             AtomicDataDict.FORCE_KEY in model.irreps_out
#             or AtomicDataDict.PARTIAL_FORCE_KEY in model.irreps_out
#     ):
#         raise ValueError("This model already has forces outputs.")
#     return PartialForceOutputModule(func=model)
#
#
# def StressForceOutput(model: GraphModuleMixin) -> StressOutputModule:
#     r"""Add forces and stresses to a model that predicts energy.
#
#     Args:
#         model: the model to wrap. Must have ``AtomicDataDict.TOTAL_ENERGY_KEY`` as an output.
#
#     Returns:
#         A ``StressOutput`` wrapping ``model``.
#     """
#     if (
#             AtomicDataDict.FORCE_KEY in model.irreps_out
#             or AtomicDataDict.STRESS_KEY in model.irreps_out
#     ):
#         raise ValueError("This model already has forces or stress outputs.")
#     return StressOutputModule(func=model)



import torch
import torch.nn as nn
from torch_geometric.data import Data
from torch_geometric.nn import aggr
from torch_scatter import scatter


class Stress(nn.Module):
    """Stress net"""

    def __init__(self, input_dim=3, readout="mean"):
        super(Stress, self).__init__()

        self.mlp0 = nn.Sequential(nn.Linear(input_dim + 3 + 81, 128, bias=True), nn.ReLU())

        self.bm = nn.BatchNorm1d(input_dim + 3 + 81)

        self.mlp2 = nn.Sequential(
            nn.Linear(128, 64), nn.ReLU(),
            nn.Linear(64, 64))

        self.readout = readout

        if self.readout == "set2set":
            self.set2set = aggr.Set2Set(64, processing_steps=2)

            self.mlp3 = nn.Sequential(
                nn.Linear(64 * 2, 32), nn.ReLU(),
                nn.Linear(32, 6))
        else:
            self.mlp3 = nn.Sequential(
                nn.Linear(64, 32), nn.ReLU(),
                nn.Linear(32, 6))

    def forward(self, atom_prop, data):
        """
        # atom_prop size (n_atoms, 3)
        # pos (n_atoms, 3)
        # cell (n_sturcture, 3, 3)
        """

        pos = data.pos
        cell = data.cell
        cell_ravel = cell.view(-1, 9)
        cell_ravel = (cell_ravel.unsqueeze(1) * cell_ravel.unsqueeze(-1)).view(-1, 81)
        cell_ravel = cell_ravel[data.batch]

        if not hasattr(data, "frac_pos"):
            cell_1 = torch.inverse(cell)
            frac_pos = torch.bmm(pos.unsqueeze(1), cell_1[data.batch], ).squeeze(1)
        else:
            frac_pos = data.frac_pos

        h = torch.cat((atom_prop, frac_pos, cell_ravel), dim=1)

        h = self.bm(h)
        h = self.mlp0(h)
        h = self.mlp2(h)

        if self.readout == "set2set":
            h = self.set2set(h, data.batch)
            stress = self.mlp3(h)
        else:
            h = scatter(h, data.batch, dim=0, reduce=self.readout)
            stress = self.mlp3(h)

        return stress


# class StressOutput(torch.nn.Module):
#     r"""Compute stress (and forces) using autograd of an energy model.
#
#     See:
#         Knuth et. al. Comput. Phys. Commun 190, 33-50, 2015
#         https://pure.mpg.de/rest/items/item_2085135_9/component/file_2156800/content
#
#     Args:
#         func: the energy model to wrap
#     """
#
#     def __init__(self, func):
#         super().__init__()
#
#         self.do_forces = True
#
#         self.func = func
#
#     def forward(self, data: Data):
#         batch = data.batch
#         num_batch = int(batch.max().cpu().item()) + 1
#         pos = data.pos
#
#         orig_cell = data.cell
#         # Make the cell per-batch
#         cell = orig_cell.view(-1, 3, 3).expand(num_batch, 3, 3)
#
#         displacement = torch.zeros(
#             (num_batch, 3, 3),
#             dtype=pos.dtype,
#             device=pos.device,
#         )
#         displacement.requires_grad_(True)
#         data["_displacement"] = displacement
#         # in the above paper, the infinitesimal distortion is *symmetric*
#         # so we symmetrize the displacement before applying it to
#         # the positions/cell
#         # This is not strictly necessary (reasoning thanks to Mario):
#         # the displacement's asymmetric 1o term corresponds to an
#         # infinitesimal rotation, which should not affect the final
#         # output (invariance).
#         # That said, due to numerical error, this will never be
#         # exactly true. So, we symmetrize the deformation to
#         # take advantage of this understanding and not rely on
#         # the invariance here:
#         symmetric_displacement = 0.5 * (displacement + displacement.transpose(-1, -2))
#         did_pos_req_grad: bool = pos.requires_grad
#         pos.requires_grad_(True)
#         # bmm is natom in batch
#         data["pos"] = pos + torch.bmm(
#             pos.unsqueeze(-2), symmetric_displacement[batch]
#         ).squeeze(-2)
#         # we only displace the cell if we have one:
#
#         data["cell"] = cell + torch.bmm(
#             cell, symmetric_displacement
#         )
#
#         # Call model and get gradients
#         energy = self.func(data)
#
#         grads = torch.autograd.grad(
#             [energy.sum()],
#             [pos, data["_displacement"]],
#             create_graph=self.training,  # needed to allow gradients of this output during training
#         )
#
#         # Put negative sign on forces
#         forces = grads[0]
#         forces = torch.neg(forces)
#
#         # data['forces'] = forces
#
#         # Store virial
#         virial = grads[1]
#
#         volume = torch.einsum(
#             "zi,zi->z",
#             cell[:, 0, :],
#             torch.cross(cell[:, 1, :], cell[:, 2, :], dim=1),
#         ).unsqueeze(-1)
#         stress = virial / volume.view(-1, 1, 1)
#
#         data["cell"] = orig_cell
#
#         # data["stress"] = stress
#
#         virial = torch.neg(virial)
#
#         # data["virial"] = virial
#
#         # Remove helper
#         del data["_displacement"]
#         if not did_pos_req_grad:
#             # don't give later modules one that does
#             pos.requires_grad_(False)
#
#         return energy, forces, stress, virial

# class Stress(nn.Module):
#     """Stress net"""
#     def __init__(self, input_dim=3):
#         super(Stress, self).__init__()
#
#         self.mlp0 = nn.Linear(input_dim, 32)
#
#         self.mlp1 = nn.Linear(3, 3, bias=True)
#         self.mlp1n = nn.Linear(3, 3, bias=True)
#
#         self.mlp2 = nn.Sequential(nn.Linear(87+32, 32), nn.ReLU(),
#                                   nn.Linear(32, 16), nn.ReLU(),
#                                   nn.Linear(16, 6), )
#         self.reset_parameters()
#         self.readout = "set2set"
#         self.readout = "mean"
#
#         if self.readout == "set2set":
#             self.set2set = aggr.Set2Set(6, processing_steps=2)
#             self.lin1 = torch.nn.Linear(2 * 6, 6)
#         else:
#             self.lin1 = torch.nn.Linear(6, 6)
#
#     def reset_parameters(self) -> None:
#         nn.init.eye_(self.mlp1.weight)
#         nn.init.eye_(self.mlp1n.weight)
#         self.mlp1.bias.data.fill_(0)
#         self.mlp1n.bias.data.fill_(0)
#
#         for m in self.mlp2:
#             if isinstance(m, nn.Linear):
#                 nn.init.xavier_uniform_(m.weight)
#                 m.bias.data.fill_(0)
#
#     def forward(self, atom_prop, data):
#         """
#         # atom_prop size (n_atoms, 3)
#         # pos (n_atoms, 3)
#         # cell (n_sturcture, 3, 3)
#         """
#
#         # atom_prop = torch.rand_like(data.pos) # for test
#
#         pos = data.pos
#         cell = data.cell
#         cell_ravel = cell.view(-1, 9)
#         cell_ravel = (cell_ravel.unsqueeze(1) * cell_ravel.unsqueeze(-1)).view(-1, 81)
#         cell_ravel = cell_ravel[data.batch]
#
#         if not hasattr(data,"frac_pos"):
#             cell_1 = torch.inverse(cell)
#             frac_pos = torch.bmm(pos.unsqueeze(1), cell_1[data.batch], ).squeeze(1)
#         else:
#             frac_pos = data.frac_pos
#
#         threshold = frac_pos - torch.floor(frac_pos) - 0.5
#
#         # threshold.requires_grad=False
#
#         atom_prop = self.mlp0(atom_prop)
#
#         threshold1 = F.relu(self.mlp1(threshold))
#         threshold2 = F.relu(self.mlp1n(-threshold))
#
#         h = torch.cat((atom_prop, threshold1, threshold2, cell_ravel), dim=1)
#
#         h = self.mlp2(h)
#
#         if self.readout == "set2set":
#             h = F.relu(self.set2set(h, data.batch))
#
#         else:
#             h = F.relu(scatter(h, data.batch, dim=0, reduce=self.readout))
#
#         stress = self.lin1(h)
#
#         return stress



"""
copy from zhengl0217:Perovskite-Electronic-Structure-Feature-Python
https://github.com/zhengl0217/Perovskite-Electronic-Structure-Feature-Python/blob/master/compositional_descriptor.py
"""
import warnings

import numpy as np

from pymatgen.core import Element

try:
    import featurebox
    from featurebox.featurizers.atom.mapper import AtomPymatgenPropMap
    from featurebox.featurizers.state.statistics import WeightedAverage

except ImportError as e:
    warnings.warn("please `pip install featurebox` for more tools", ImportWarning)
    raise e


def T_f(rA, rB, rX):
    """
    公差系数。

    Parameters
    ----------
    rA:
        A位离子半径
    rB:
        B位离子半径
    rX:
        X位原子的离子半径
    """
    t_f = (rA + rX) / (np.sqrt(2) * (rB + rX))
    return t_f


def T_f_C_organic(rA, rB, rX, rXh):
    """
    公差系数,C-site 有机。

    Parameters
    ----------
    rA:
        A位离子半径
    rB:
        B位离子半径
    rX:
        X位原子的离子半径
    """
    t_f = (rA + rX) / (np.sqrt(2) * (rB + 0.5*rXh))
    return t_f

def O_f(rB, rX):
    """
    八面体因子

    Parameters
    ----------
    rB:
        B位离子半径
    rX:
        X位原子的离子半径

    Returns
    -------

    """
    # rB,rX分别表示B位和X位原子的离子半径
    # Of 八面体因子
    o_f = rB / rX

    return o_f


def r_sp(s, p):
    """
    s和p轨道半径之和(B, X)

    Parameters
    ----------
    s:
        s 轨道半径
    p:
        轨道半径

    """
    r_sp2 = s + p
    return r_sp2


def elemental_feature(A, B, C):
    """Not for molecule"""
    data = [A, B, C]
    for i in range(len(data)):
        if isinstance(data[i], dict):
            data[i] = [data[i], ]
    A, B, C = data

    func_map = [
        "common_oxidation_states"
        "average_ionic_radius",
        "average_cationic_radius",
    ]
    appma = AtomPymatgenPropMap(prop_name=func_map, search_tp="name")
    wa = WeightedAverage(appma, n_jobs=1, return_type="df")
    wa.set_feature_labels(["ave_A_" + i for i in appma.feature_labels])
    dA2 = wa.fit_transform(A)
    wa.set_feature_labels(["ave_B_" + i for i in appma.feature_labels])
    dB2 = wa.fit_transform(B)
    wa.set_feature_labels(["ave_C_" + i for i in appma.feature_labels])
    dC2 = wa.fit_transform(C)

    fff17 = dA2["ave_A_average_ionic_radius"]
    fff18 = dB2["ave_B_average_ionic_radius"]
    fff19 = dC2["ave_C_average_ionic_radius"]
    # Octahedral factor
    # Tolerance factor
    fff_22 = O_f(fff17, fff18)
    fff_23 = T_f(fff17, fff18, fff19)

    # A/B ion oxidation state
    ff_24 = (dA2["ave_A_common_oxidation_states"][0]) / (dB2["ave_B_common_oxidation_states"][0])
    # ionic_radius ratios
    ionic_ration_AO = fff17 / fff19
    ionic_ration_BO = fff18 / fff19

    return fff_22, fff_23, ff_24, ionic_ration_AO, ionic_ration_BO


def generalized_mean(x, p, N):
    """Generalized mean function capture the mean value of atomic properties by considering the ratio of each element in the structure.
    Args:
       x (np.ndarray): array of atomic properties for each atom in the structure.
       p (int): power parameter, e.g., harmonic mean (-1), geometric mean(0), arithmetic mean(1), quadratic mean(2).
       N (int): total number of atoms in the structure.
    Returns:
       float: generalized mean value.
    """
    if p != 0:
        D = 1 / (N)
        out = (D * sum(x ** p)) ** (1 / p)
    else:
        D = 1 / (N)
        out = np.exp(D * sum(np.log(x)))
    return out


def geometric_descriptor(element_dict):
    """Extract geometric mean of the atomic properties in the perovskite structure.
    Args:
       element_dict (dict): element frequency library in a perovskite structure, e.g., {'La': 4, 'Ba': 4, 'Co': 8, 'O': 24}
    Returns:
       np.ndarray: geometric based descriptor, including atomic_radius,  mendeleev number', common_oxidation_states, Pauling electronegativity, thermal_conductivity, average_ionic_radius, atomic_orbitals.
    """
    # encode the orbital types
    category = {'s': 1, 'p': 2, 'd': 3, 'f': 4}
    # total number of atoms in a perovskite structure
    N = sum(element_dict.values())
    # obtain array of atomic properties for each element type
    atomic_number_list = []
    atomic_mass_list = []
    atomic_radius_list = []
    mendeleev_no_list = []
    common_oxidation_states_list = []
    Pauling_electronegativity_list = []
    row_list = []
    group_list = []
    block_list = []
    thermal_conductivity_list = []
    boiling_point_list = []
    melting_point_list = []
    average_ionic_radius_list = []
    molar_volume_list = []
    atomic_orbitals_list = []
    for item in element_dict:
        # extract atomic property from pymatgen
        ele = Element(item)
        atomic_number = ele.Z
        atomic_mass = float(str(ele.atomic_mass)[:-4])
        atomic_radius = float(str(ele.atomic_radius)[:-4])
        mendeleev_no = ele.mendeleev_no
        common_oxidation_states = ele.common_oxidation_states[0]
        Pauling_electronegativity = ele.X
        row = ele.row
        group = ele.group
        block = ele.block
        thermal_conductivity = float(str(ele.thermal_conductivity)[:-12])
        boiling_point = float(str(ele.boiling_point)[: -2])
        melting_point = float(str(ele.melting_point)[: -2])
        average_ionic_radius = float(str(ele.average_ionic_radius)[:-4])
        molar_volume = float(str(ele.molar_volume)[: -5])
        if '6s' in ele.atomic_orbitals.keys():
            atomic_orbitals = ele.atomic_orbitals['6s']
        elif '4s' in ele.atomic_orbitals.keys():
            atomic_orbitals = ele.atomic_orbitals['4s']
        else:
            atomic_orbitals = ele.atomic_orbitals['2s']
        # calculate the array of atomic properties for all atoms
        atomic_number_list += [atomic_number] * element_dict[item]
        atomic_mass_list += [atomic_mass] * element_dict[item]
        atomic_radius_list += [atomic_radius] * element_dict[item]
        mendeleev_no_list += [mendeleev_no] * element_dict[item]
        common_oxidation_states_list += [common_oxidation_states] * element_dict[item]
        Pauling_electronegativity_list += [Pauling_electronegativity] * element_dict[item]
        row_list += [row] * element_dict[item]
        group_list += [group] * element_dict[item]
        block_list += [category[block]] * element_dict[item]
        thermal_conductivity_list += [thermal_conductivity] * element_dict[item]
        boiling_point_list += [boiling_point] * element_dict[item]
        melting_point_list += [melting_point] * element_dict[item]
        average_ionic_radius_list += [average_ionic_radius] * element_dict[item]
        molar_volume_list += [molar_volume] * element_dict[item]
        atomic_orbitals_list += [atomic_orbitals] * element_dict[item]
    return [generalized_mean(np.array(atomic_number_list), 1, N)] + [
        generalized_mean(np.array(atomic_radius_list), 1, N)] + [
               generalized_mean(np.array(mendeleev_no_list), 1, N)] + [
               generalized_mean(np.array(common_oxidation_states_list), 1, N)] + [
               generalized_mean(np.array(Pauling_electronegativity_list), 1, N)] + [
               generalized_mean(np.array(thermal_conductivity_list), 1, N)] + [
               generalized_mean(np.array(average_ionic_radius_list), 1, N)] + [
               generalized_mean(np.array(atomic_orbitals_list), 1, N)]


if __name__ == "__main__":
    print('LaBaCo2O6',
          geometric_descriptor({'La': 4, 'Ba': 4, 'Co': 8, 'O': 24}))



import queue
from collections import deque

import numpy as np
import warnings

from scipy.optimize import minimize
from ase.parallel import world
from ase.io.jsonio import write_json
from ase.optimize.optimize import Optimizer
from ase.optimize.gpmin.kernel import SquaredExponential
from ase.optimize.gpmin.prior import ConstantPrior

from .basegpmin import BaseGPMin
from .gp import GaussianProcess


class StrengthenGPMin(BaseGPMin):
    def __init__(self, atoms, update_hyperparams=True, batch_size=10, **kwargs):
        super(StrengthenGPMin, self).__init__(atoms, update_hyperparams= update_hyperparams,
                                              batch_size=batch_size, **kwargs)

    def step(self, f=None):
        atoms = self.atoms
        if f is None:
            f = atoms.get_forces()

        fc = self.force_consistent
        r0 = self.atoms.get_positions().reshape(-1)
        e0 = self.atoms.get_potential_energy(force_consistent=fc)
        self.update(r0, e0, f)

        r1 = self.relax_model(r0)
        self.atoms.set_positions(r1.reshape(-1, 3))
        e1 = self.atoms.get_potential_energy(force_consistent=fc)
        f1 = self.atoms.get_forces()
        self.function_calls += 1
        self.force_calls += 1
        count = 0

        while e1 >= e0:
            self.update(r1, e1, f1)

            random_factor = max(0.01 * 1 / (self.nsteps + 1),
                                0.001)  # decrease the degree of position change with step.

            r_ = self.resample_r(20, random_factor=random_factor)

            res = self.predict_mean_std_all(r_)

            ei = self.ei_balance(res)

            index_ = np.argsort(-ei)[:3]
            r_add = r_[index_]

            r_add, e_add, f_add = self.r_to_e_f(r_add)
            self.update_batch(r_add, e_add, f_add)  # train gp model

            r1 = self.relax_model(r0)
            self.atoms.set_positions(r1.reshape(-1, 3))
            e1 = self.atoms.get_potential_energy(force_consistent=self.force_consistent)
            f1 = self.atoms.get_forces()

            self.function_calls += 1
            self.force_calls += 1

            if self.converged(f1):
                break

            count += 1
            if count == 5:
                warnings.warn("A descent model could not be built")
                break

        self.dump()











    


"""
From NequlIP

Keys for dictionaries/AtomicData objects.

This is a seperate module to compensate for a TorchScript bug
that can only recognize constants when they are accessed as attributes of an imported module.
"""
import sys
from typing import List

if sys.version_info[1] >= 8:
    from typing import Final
else:
    from typing_extensions import Final

# == Define allowed keys as constants ==
# The positions of the atoms in the system
POSITIONS_KEY: Final[str] = "pos"
# The [2, n_edge] index tensor giving center -> neighbor relations
EDGE_INDEX_KEY: Final[str] = "edge_index"
# A [n_edge, 3] tensor of how many periodic cells each edge crosses in each cell vector
EDGE_CELL_SHIFT_KEY: Final[str] = "cell_offsets"
# [n_batch, 3, 3] or [3, 3] tensor where rows are the cell vectors
CELL_KEY: Final[str] = "cell"
# [n_batch, 3] bool tensor
PBC_KEY: Final[str] = "pbc"
# [n_atom, 1] long tensor
ATOMIC_NUMBERS_KEY: Final[str] = "z"
# [n_atom, 1] long tensor
ATOM_TYPE_KEY: Final[str] = "zt"

BASIC_STRUCTURE_KEYS: Final[List[str]] = [
    POSITIONS_KEY,
    EDGE_INDEX_KEY,
    EDGE_CELL_SHIFT_KEY,
    CELL_KEY,
    PBC_KEY,
    ATOM_TYPE_KEY,
    ATOMIC_NUMBERS_KEY,
]

# A [n_edge, 3] tensor of displacement vectors associated to edges
EDGE_VECTORS_KEY: Final[str] = "edge_vec"
# A [n_edge] tensor of the lengths of EDGE_VECTORS
EDGE_LENGTH_KEY: Final[str] = "edge_weight"
# [n_edge, dim] (possibly equivariant) attributes of each edge
EDGE_ATTRS_KEY: Final[str] = "edge_attr"
# [n_edge, dim] invariant embedding of the edges
EDGE_EMBEDDING_KEY: Final[str] = "edge_embedding"
EDGE_FEATURES_KEY: Final[str] = "edge_features"

NODE_FEATURES_KEY: Final[str] = "x"
NODE_ATTRS_KEY: Final[str] = "node_attr"

PER_ATOM_ENERGY_KEY: Final[str] = "atomic_energy"
TOTAL_ENERGY_KEY: Final[str] = "energy"
FORCE_KEY: Final[str] = "forces"
PARTIAL_FORCE_KEY: Final[str] = "partial_force"
STRESS_KEY: Final[str] = "stress"
VIRIAL_KEY: Final[str] = "virial"

ALL_ENERGY_KEYS: Final[List[str]] = [
    PER_ATOM_ENERGY_KEY,
    TOTAL_ENERGY_KEY,
    FORCE_KEY,
    PARTIAL_FORCE_KEY,
    STRESS_KEY,
    VIRIAL_KEY,
]

BATCH_KEY: Final[str] = "batch"

# Make a list of allowed keys
ALLOWED_KEYS: List[str] = [
    getattr(sys.modules[__name__], k)
    for k in sys.modules[__name__].__dict__.keys()
    if k.endswith("_KEY")
]



"""
Examples:
    >>> data = pd.read_pickle("st_and_energy.pkl_pd")
    >>> ks, (structure, energy), k2s = unpack(data)
    >>> data1 = pack(ks, (structure, energy), k2s)
"""

from typing import List, Sequence, Union

import numpy as np
import pandas as pd
from pymatgen.core import Structure
from tqdm import tqdm


def pack(ks: Union[Sequence, np.ndarray], vs: Union[Sequence, np.ndarray],
         k2s: Union[Sequence, np.ndarray] = None, trans: bool = True):
    """Pack the array to dict."""

    res4 = pd.DataFrame.from_dict({i: j for i, j in zip(ks, vs)})

    if k2s is None:
        pass
    else:
        res4.index = k2s

    if trans:
        res4 = res4.T

    data = res4.to_dict()
    return data


def unpack(data: dict, trans=True):
    """Unpack the dict to names and value."""
    res = pd.DataFrame.from_dict(data)
    if trans:
        res = res.T
    ks = res.columns.values
    vs = res.values
    vs = [vs[:, i] for i in range(vs.shape[1])]
    k2s = res.index.values
    return ks, vs, k2s


def unpack2dict(data: dict, trans=True):
    """Un-squeeze the 2 layer dict to simple data."""
    ks, vs, k2s = unpack(data, trans=trans)
    data_new = {ksi: vsi for ksi, vsi in zip(ks, vs)}
    return data_new, k2s


def couple_org(structures: List[Structure], energies: Sequence = None, **kwargs):
    """Group (*,*H) structure by formula."""
    if energies is None:
        energies = [np.nan] * len(structures)
    kwargs2 = {}
    kwargs2.update({"energies": energies})
    kwargs2.update(kwargs)

    # name = list(kwargs.keys())
    values = list(kwargs2.values())

    ans = [set(si.composition.formula.split(" ")) for si in structures]
    l_ans = len(ans)
    st_dict = [(si, ni) for si, ni in zip(structures, range(l_ans))]

    couple = [[] for _ in range(1 + len(values))]

    for i in tqdm(range(l_ans)):
        for j in range(i + 1, l_ans):
            if ans[i].issubset(ans[j]) and len(ans[j] - ans[i]) == 1:
                st1 = st_dict[i][0]
                st2 = st_dict[j][0]
                c2 = st2.cart_coords[:-1]
                c1 = st1.cart_coords
                if len(st2.atomic_numbers) - len(st1.atomic_numbers) == 1:
                    if np.sum(np.abs(c2 - c1)) / len(c1) < 100 and np.sum(np.abs(c2 - c1)) / len(c1) > 1:  # ????
                        # diff = np.sum(np.abs(c2 - c1)) / len(c1) # structure difference
                        couple[0].extend((st_dict[i][0], st_dict[j][0]))
                        for ni, vi in enumerate(values):
                            couple[ni + 1].extend((vi[i], vi[j]))

    return couple


if __name__ == "__main__":
    data = pd.read_pickle("st_and_energy.pkl_pd")
    ks, (structure, energy), k2s = unpack(data)
    data1 = pack(ks, (structure, energy), k2s)



# -*- coding: utf-8 -*-
# @Time  : 2023/5/5 13:21
# @Author : boliqq07
# @Software: PyCharm
# @License: MIT License

from ase import Atoms


def molecule_generator(molecule_data):
    for i, j in molecule_data.items():
        if isinstance(j["atom"], list):
            yield i, Atoms(symbols=j["atom"])
        else:
            yield i, j["atom"]


"""
From ocp
"""

from typing import List

import math
import torch
import torch.nn as nn


class Sine(nn.Module):
    def __init__(self, w0: float = 30.0):
        super(Sine, self).__init__()
        self.w0 = w0

    def forward(self, x: torch.Tensor) -> torch.Tensor:
        return torch.sin(self.w0 * x)


class SIREN(nn.Module):
    def __init__(
            self,
            layers: List[int],
            in_features: int,
            out_features: int,
            w0: float = 30.0,
            initializer: str = "siren",
            c: float = 6,
    ):

        super(SIREN, self).__init__()
        self.layers = [nn.Linear(in_features, layers[0]), Sine(w0=w0)]

        for index in range(len(layers) - 1):
            self.layers.extend(
                [nn.Linear(layers[index], layers[index + 1]), Sine(w0=1)]
            )

        self.layers.append(nn.Linear(layers[-1], out_features))
        self.network = nn.Sequential(*self.layers)

        if initializer is not None and initializer == "siren":
            for m in self.network:
                if isinstance(m, nn.Linear):
                    num_input = float(m.weight.size(-1))
                    with torch.no_grad():
                        m.weight.uniform_(
                            -math.sqrt(6.0 / num_input),
                            math.sqrt(6.0 / num_input),
                        )

    def forward(self, X):
        return self.network(X)


class SINESmearing(nn.Module):
    def __init__(self, in_features, num_freqs=40, use_cosine=False):

        super(SINESmearing, self).__init__()

        self.num_freqs = num_freqs
        self.out_dim = in_features * self.num_freqs
        self.use_cosine = use_cosine

        freq = torch.arange(num_freqs).float()
        freq = torch.pow(torch.ones_like(freq) * 1.1, freq)
        self.freq_filter = nn.Parameter(
            freq.view(-1, 1).repeat(1, in_features).view(1, -1),
            requires_grad=False,
        )

    def forward(self, x):
        x = x.repeat(1, self.num_freqs)
        x = x * self.freq_filter

        if self.use_cosine:
            return torch.cos(x)
        else:
            return torch.sin(x)


class GaussianSmearing(nn.Module):
    def __init__(self, in_features, start=0, end=1, num_freqs=50):
        super(GaussianSmearing, self).__init__()
        self.num_freqs = num_freqs
        offset = torch.linspace(start, end, num_freqs)
        self.coeff = -0.5 / (offset[1] - offset[0]).item() ** 2
        self.offset = nn.Parameter(
            offset.view(-1, 1).repeat(1, in_features).view(1, -1),
            requires_grad=False,
        )

    def forward(self, x):
        x = x.repeat(1, self.num_freqs)
        x = x - self.offset
        return torch.exp(self.coeff * torch.pow(x, 2))


class FourierSmearing(nn.Module):
    def __init__(self, in_features, num_freqs=40, use_cosine=False):

        super(FourierSmearing, self).__init__()

        self.num_freqs = num_freqs
        self.out_dim = in_features * self.num_freqs
        self.use_cosine = use_cosine

        freq = torch.arange(num_freqs).to(torch.float32)
        self.freq_filter = nn.Parameter(
            freq.view(-1, 1).repeat(1, in_features).view(1, -1),
            requires_grad=False,
        )

    def forward(self, x):
        x = x.repeat(1, self.num_freqs)
        x = x * self.freq_filter

        if self.use_cosine:
            return torch.cos(x)
        else:
            return torch.sin(x)

# class Basis(nn.Module):
#     def __init__(
#             self,
#             in_features,
#             num_freqs=50,
#             basis_type="powersine",
#             act="ssp",
#             sph=None,
#     ):
#         super(Basis, self).__init__()
#
#         self.num_freqs = num_freqs
#         self.basis_type = basis_type
#
#         if basis_type == "powersine":
#             self.smearing = SINESmearing(in_features, num_freqs)
#             self.out_dim = in_features * num_freqs
#         elif basis_type == "powercosine":
#             self.smearing = SINESmearing(
#                 in_features, num_freqs, use_cosine=True
#             )
#             self.out_dim = in_features * num_freqs
#         elif basis_type == "fouriersine":
#             self.smearing = FourierSmearing(in_features, num_freqs)
#             self.out_dim = in_features * num_freqs
#         elif basis_type == "gauss":
#             self.smearing = GaussianSmearing(
#                 in_features, start=0, end=1, num_freqs=num_freqs
#             )
#             self.out_dim = in_features * num_freqs
#         elif basis_type == "linact":
#             self.smearing = torch.nn.Sequential(
#                 torch.nn.Linear(in_features, num_freqs * in_features), Act(act)
#             )
#             self.out_dim = in_features * num_freqs
#         elif basis_type == "raw" or basis_type == "rawcat":
#             self.out_dim = in_features
#         elif "sph" in basis_type:
#             # by default, we use sine function to encode distance
#             # sph must be given here
#             assert sph is not None
#             # assumes the first three columns are normalizaed xyz
#             # the rest of the columns are distances
#             if "cat" in basis_type:
#                 # concatenate
#                 self.smearing_sine = SINESmearing(in_features - 3, num_freqs)
#                 self.out_dim = sph.out_dim + (in_features - 3) * num_freqs
#             elif "mul" in basis_type:
#                 self.smearing_sine = SINESmearing(in_features - 3, num_freqs)
#                 self.lin = torch.nn.Linear(
#                     self.smearing_sine.out_dim, in_features - 3
#                 )
#                 self.out_dim = (in_features - 3) * sph.out_dim
#             elif "m40" in basis_type:
#                 dim = 40
#                 self.smearing_sine = SINESmearing(in_features - 3, num_freqs)
#                 self.lin = torch.nn.Linear(
#                     self.smearing_sine.out_dim, dim
#                 )  # make the output dimensionality comparable.
#                 self.out_dim = dim * sph.out_dim
#             elif "nosine" in basis_type:
#                 # does not use sine smearing for encoding distance
#                 self.out_dim = (in_features - 3) * sph.out_dim
#             else:
#                 raise ValueError(
#                     "cat or mul not specified for spherical harnomics."
#                 )
#         else:
#             raise RuntimeError("Undefined basis type.")
#
#     def forward(self, x, edge_attr_sph=None):
#         if "sph" in self.basis_type:
#             if "nosine" not in self.basis_type:
#                 x_sine = self.smearing_sine(
#                     x[:, 3:]
#                 )  # the first three features correspond to edge_vec_normalized, so we ignore
#                 if "cat" in self.basis_type:
#                     # just concatenate spherical edge feature and sined node features
#                     return torch.cat([edge_attr_sph, x_sine], dim=1)
#                 elif "mul" in self.basis_type or "m40" in self.basis_type:
#                     # multiply sined node features into spherical edge feature (inspired by theory in spherical harmonics)
#                     r = self.lin(x_sine)
#                     outer = torch.einsum("ik,ij->ikj", edge_attr_sph, r)
#                     return torch.flatten(outer, start_dim=1)
#                 else:
#                     raise RuntimeError(
#                         f"Unknown basis type called {self.basis_type}"
#                     )
#             else:
#                 outer = torch.einsum("ik,ij->ikj", edge_attr_sph, x[:, 3:])
#                 return torch.flatten(outer, start_dim=1)
#
#         elif "raw" in self.basis_type:
#             # do nothing, just return node features
#             pass
#         else:
#             x = self.smearing(x)
#         return x



from typing import Union

import torch
from e3nn import o3
from e3nn.util.jit import compile_mode

from hpc2ml.data.nequlip_support import AtomicDataDict
from hpc2ml.data.nequlip_support.AtomicDataDict import ST
from .._graph_mixin import GraphModuleMixin
from ..cutoffs import PolynomialCutoff
from ..radial_basis import BesselBasis


@compile_mode("script")
class SphericalHarmonicEdgeAttrs(GraphModuleMixin, torch.nn.Module):
    """Construct edge attrs as spherical harmonic projections of edge vectors.

    Parameters follow ``e3nn.o3.spherical_harmonics``.

    Args:
        irreps_edge_sh (int, str, or o3.Irreps): if int, will be treated as lmax for o3.Irreps.spherical_harmonics(lmax)
        edge_sh_normalization (str): the normalization scheme to use
        edge_sh_normalize (bool, default: True): whether to normalize the spherical harmonics
        out_field (str, default: AtomicDataDict.EDGE_ATTRS_KEY: data/irreps in_field
    """

    out_field: str

    def __init__(
            self,
            irreps_in=None,
            out_field: str = AtomicDataDict.EDGE_ATTRS_KEY,
            irreps_edge_sh: Union[int, str, o3.Irreps] = 2,
            edge_sh_normalization: str = "component",
            edge_sh_normalize: bool = True,

    ):
        super().__init__()
        self.out_field = out_field
        # irreps_edge_sh is the irreps_edge_sh

        if isinstance(irreps_edge_sh, int):
            self.irreps_edge_sh = o3.Irreps.spherical_harmonics(irreps_edge_sh)
        else:
            self.irreps_edge_sh = o3.Irreps(irreps_edge_sh)
        self._init_irreps(
            irreps_in=irreps_in,
            irreps_out={out_field: self.irreps_edge_sh},
        )
        self.sh = o3.SphericalHarmonics(
            self.irreps_edge_sh, edge_sh_normalize, edge_sh_normalization
        )

    def forward(self, data: ST) -> ST:
        data = AtomicDataDict.with_edge_vectors(data, with_lengths=False)
        edge_vec = data[AtomicDataDict.EDGE_VECTORS_KEY]
        edge_sh = self.sh(edge_vec)
        data[self.out_field] = edge_sh
        return data


@compile_mode("script")
class RadialBasisEdgeEncoding(GraphModuleMixin, torch.nn.Module):
    out_field: str

    def __init__(
            self,
            irreps_in=None,
            out_field: str = AtomicDataDict.EDGE_EMBEDDING_KEY,
            basis=BesselBasis, cutoff=PolynomialCutoff, basis_kwargs=None, cutoff_kwargs=None, ):
        super().__init__()

        if cutoff_kwargs is None:
            cutoff_kwargs = {}
        if basis_kwargs is None:
            basis_kwargs = {}

        self.basis = basis(**basis_kwargs)
        self.cutoff = cutoff(**cutoff_kwargs)
        self.out_field = out_field
        self._init_irreps(
            irreps_in=irreps_in,
            irreps_out={self.out_field: o3.Irreps([(self.basis.num_basis, (0, 1))])},
        )

    def forward(self, data: ST) -> ST:
        data = AtomicDataDict.with_edge_vectors(data, with_lengths=True)
        edge_length = data[AtomicDataDict.EDGE_LENGTH_KEY]
        edge_length_embedded = (
                self.basis(edge_length) * self.cutoff(edge_length)[:, None]
        )
        data[self.out_field] = edge_length_embedded
        return data



"""
From NequlIP

nequip.data.jit: TorchScript functions for dealing with AtomicData.

These TorchScript functions operate on ``Dict[str, torch.Tensor]`` representations
of the ``AtomicData`` class which are produced by ``AtomicData.to_AtomicDataDict()``.

# Also import the module to use in TorchScript, this is a hack to avoid bug:
# https://github.com/pytorch/pytorch/issues/52312

from .nequlip_support._keys import *

Authors: Albert Musaelian
"""

import torch
import torch.jit

from ._keys import *  # this is import
from . import _keys
from ._irrreps_support import _fix_irreps_dict, _irreps_compatible  # this is import
from .typing import _SI, _ST

# alias Types
SI = _SI
ST = _ST


def validate_keys(keys, graph_required=True):
    # Validate combinations
    if graph_required:
        if not (_keys.POSITIONS_KEY in keys and _keys.EDGE_INDEX_KEY in keys):
            raise KeyError("At least pos and edge_index must be supplied")
    if _keys.EDGE_CELL_SHIFT_KEY in keys and "cell" not in keys:
        raise ValueError("If `edge_cell_shift` given, `cell` must be given.")


@torch.jit.script
def with_edge_vectors(data: ST, with_lengths: bool = True) -> ST:
    """Compute the edge displacement vectors for a graph.

    If ``data.pos.requires_grad`` and/or ``data.cell.requires_grad``, this
    method will return edge vectors correctly connected in the autograd graph.

    Returns:
        Tensor [n_edges, 3] edge displacement vectors
    """
    if _keys.EDGE_VECTORS_KEY in data:
        if with_lengths and _keys.EDGE_LENGTH_KEY not in data:
            data[_keys.EDGE_LENGTH_KEY] = torch.linalg.norm(
                data[_keys.EDGE_VECTORS_KEY], dim=-1
            )
        return data
    else:
        # Build it dynamically
        # Note that this is
        # (1) backwardable, because everything (pos, cell, shifts)
        #     is Tensors.
        # (2) works on a Batch constructed from AtomicData
        pos = data[_keys.POSITIONS_KEY]
        edge_index = data[_keys.EDGE_INDEX_KEY]
        edge_vec = pos[edge_index[1]] - pos[edge_index[0]]
        if _keys.CELL_KEY in data:
            # ^ note that to save time we don't check that the edge_cell_shifts are trivial if no cell is provided; we just assume they are either not present or all zero.
            # -1 gives a batch dim no matter what
            cell = data[_keys.CELL_KEY].view(-1, 3, 3)
            edge_cell_shift = data[_keys.EDGE_CELL_SHIFT_KEY]
            if cell.shape[0] > 1:
                batch = data[_keys.BATCH_KEY]
                # Cell has a batch dimension
                # note the ASE cell vectors as rows convention
                edge_vec = edge_vec + torch.einsum("ni,nij->nj", edge_cell_shift, cell[batch[edge_index[0]]])
                # creating an [n_edge] and [n_edge, 3, 3] tensor?
            else:
                # Cell has either no batch dimension, or a useless one,
                # so we can avoid creating the large intermediate cell tensor.
                # Note that we do NOT check that the batch array, if it is present,
                # is trivial — but this does need to be consistent.
                edge_vec = edge_vec + torch.einsum(
                    "ni,ij->nj",
                    edge_cell_shift,
                    cell.squeeze(0),  # remove batch dimension
                )
        data[_keys.EDGE_VECTORS_KEY] = edge_vec
        if with_lengths:
            data[_keys.EDGE_LENGTH_KEY] = torch.linalg.norm(edge_vec, dim=-1)
        return data


@torch.jit.script
def with_batch(data: ST) -> ST:
    """Get batch Tensor.

    If this AtomicDataPrimitive has no ``batch``, one of all zeros will be
    allocated and returned.
    """
    if _keys.BATCH_KEY in data:
        return data
    else:
        pos = data[_keys.POSITIONS_KEY]
        batch = torch.zeros(len(pos), dtype=torch.long, device=pos.device)
        data[_keys.BATCH_KEY] = batch
        return data



from typing import Optional

import math
import torch
from e3nn.math import soft_one_hot_linspace
from e3nn.util.jit import compile_mode
from torch import nn


@compile_mode("trace")
class e3nn_basis(nn.Module):
    r_max: float
    r_min: float
    e3nn_basis_name: str
    num_basis: int

    def __init__(
            self,
            r_max: float,
            r_min: Optional[float] = None,
            e3nn_basis_name: str = "gaussian",
            num_basis: int = 8,
    ):
        super().__init__()
        self.r_max = r_max
        self.r_min = r_min if r_min is not None else 0.0
        self.e3nn_basis_name = e3nn_basis_name
        self.num_basis = num_basis

    def forward(self, x: torch.Tensor) -> torch.Tensor:
        return soft_one_hot_linspace(
            x,
            start=self.r_min,
            end=self.r_max,
            number=self.num_basis,
            basis=self.e3nn_basis_name,
            cutoff=True,
        )

    def _make_tracing_inputs(self, n: int):
        return [{"forward": (torch.randn(5, 1),)} for _ in range(n)]


class BesselBasis(nn.Module):
    r_max: float
    prefactor: float

    def __init__(self, r_max, num_basis=8, trainable=True):
        r"""Radial Bessel Basis, as proposed in DimeNet: https://arxiv.org/abs/2003.03123


        Parameters
        ----------
        r_max : float
            Cutoff radius

        num_basis : int
            Number of Bessel Basis functions

        trainable : bool
            Train the :math:`n \pi` part or not.
        """
        super(BesselBasis, self).__init__()

        self.trainable = trainable
        self.num_basis = num_basis

        self.r_max = float(r_max)
        self.prefactor = 2.0 / self.r_max

        bessel_weights = (
                torch.linspace(start=1.0, end=num_basis, steps=num_basis) * math.pi
        )
        if self.trainable:
            self.bessel_weights = nn.Parameter(bessel_weights)
        else:
            self.register_buffer("bessel_weights", bessel_weights)

    def forward(self, x: torch.Tensor) -> torch.Tensor:
        """
        Evaluate Bessel Basis for input x.

        Parameters
        ----------
        x : torch.Tensor
            Input
        """
        numerator = torch.sin(self.bessel_weights * x.unsqueeze(-1) / self.r_max)

        return self.prefactor * (numerator / x.unsqueeze(-1))

# class GaussianBasis(nn.Module):
#     r_max: float

#     def __init__(self, r_max, r_min=0.0, num_basis=8, trainable=True):
#         super().__init__()

#         self.trainable = trainable
#         self.num_basis = num_basis

#         self.r_max = float(r_max)
#         self.r_min = float(r_min)

#         means = torch.linspace(self.r_min, self.r_max, self.num_basis)
#         stds = torch.full(size=means.size, fill_value=means[1] - means[0])
#         if self.trainable:
#             self.means = nn.Parameter(means)
#             self.stds = nn.Parameter(stds)
#         else:
#             self.register_buffer("means", means)
#             self.register_buffer("stds", stds)

#     def forward(self, x: torch.Tensor) -> torch.Tensor:
#         x = (x[..., None] - self.means) / self.stds
#         x = x.square().mul(-0.5).exp() / self.stds  # sqrt(2 * pi)



# -*- coding: utf-8 -*-

# @Time    : 2021/7/15 22:17
# @Email   : 986798607@qq.com
# @Software: PyCharm
# @License: BSD 3-Clause

from typing import List, Tuple

from ase import Atoms


def rotate_times(atoms: Atoms, times: int = 4, axis: Tuple[Tuple, int, str] = ((1, 0, 0),)) -> List[Atoms]:
    """Rotate atoms by 360/times along x, y, z axis"""
    atss = []
    atoms.center()
    ang = int(360 / times)
    if axis is None:
        axis = [(1, 0, 0), (0, 1, 0), (0, 0, 1)]
    if isinstance(axis, int) and axis < 3:
        axist = [[0, 0, 0]]
        axist[0][axis] = 1
        axis = axist
    for axi in axis:
        for i in range(times):
            atomc = atoms.copy()
            atomc.rotate(ang * i, v=axi)
            atss.append(atomc)

    return atss


def rotate_by_angle(atoms: Atoms, angle: int = 90, axis: Tuple[Tuple, int] = ((1, 0, 0),), center=(0, 0, 0)) -> List[
    Atoms]:
    """Rotate atoms by angle along x, y, z axis"""
    atss = []
    atoms.center()

    if axis is None:
        axis = [(1, 0, 0), (0, 1, 0), (0, 0, 1)]
    if isinstance(axis, int) and axis < 3:
        axist = [[0, 0, 0]]
        axist[0][axis] = 1
        axis = axist
    for axi in axis:
        atomc = atoms.copy()
        atomc.rotate(angle, v=axi, center=center)
        atss.append(atomc)

    return atss



import logging

import torch
from e3nn import o3
from e3nn.nn import Gate, NormActivation

from hpc2ml.data.nequlip_support import AtomicDataDict
from hpc2ml.data.nequlip_support.AtomicDataDict import ST
from ._interaction_block import InteractionBlock
from ._graph_mixin import GraphModuleMixin
from .nonlinearities import ShiftedSoftPlus


def tp_path_exists(irreps_in1, irreps_in2, ir_out):
    irreps_in1 = o3.Irreps(irreps_in1).simplify()
    irreps_in2 = o3.Irreps(irreps_in2).simplify()
    ir_out = o3.Irrep(ir_out)

    for _, ir1 in irreps_in1:
        for _, ir2 in irreps_in2:
            if ir_out in ir1 * ir2:
                return True
    return False


acts = {
    "abs": torch.abs,
    "tanh": torch.tanh,
    "ssp": ShiftedSoftPlus,
    "silu": torch.nn.functional.silu,
}


class ConvNetLayer(GraphModuleMixin, torch.nn.Module):
    """
    Args:

    """

    resnet: bool

    def __init__(
            self,
            irreps_in,
            feature_irreps_hidden,
            convolution=InteractionBlock,
            out_field=AtomicDataDict.NODE_FEATURES_KEY,
            convolution_kwargs=None,
            num_layers: int = 3,
            resnet: bool = False,
            nonlinearity_type: str = "gate",
            nonlinearity_scalars=None,
            nonlinearity_gates=None,
    ):
        super().__init__()
        assert out_field == AtomicDataDict.NODE_FEATURES_KEY

        # initialization
        if nonlinearity_gates is None:
            nonlinearity_gates = {"e": "silu", "o": "abs"}
        if nonlinearity_scalars is None:
            nonlinearity_scalars = {"e": "silu", "o": "tanh"}
        if convolution_kwargs is None:
            convolution_kwargs = {}
        assert nonlinearity_type in ("gate", "norm")
        # make the nonlin dicts from parity ints instead of convinience strs
        nonlinearity_scalars = {
            1: nonlinearity_scalars["e"],
            -1: nonlinearity_scalars["o"],
        }
        nonlinearity_gates = {
            1: nonlinearity_gates["e"],
            -1: nonlinearity_gates["o"],
        }

        self.feature_irreps_hidden = o3.Irreps(feature_irreps_hidden)
        self.resnet = resnet
        self.num_layers = num_layers

        # We'll set irreps_out later when we know them
        self._init_irreps(
            irreps_in=irreps_in,
            required_irreps_in=[AtomicDataDict.NODE_FEATURES_KEY],
        )

        edge_attr_irreps = self.irreps_in[AtomicDataDict.EDGE_ATTRS_KEY]
        irreps_layer_out_prev = self.irreps_in[AtomicDataDict.NODE_FEATURES_KEY]

        irreps_scalars = o3.Irreps(
            [
                (mul, ir)
                for mul, ir in self.feature_irreps_hidden
                if ir.l == 0
                   and tp_path_exists(irreps_layer_out_prev, edge_attr_irreps, ir)
            ]
        )

        irreps_gated = o3.Irreps(
            [
                (mul, ir)
                for mul, ir in self.feature_irreps_hidden
                if ir.l > 0
                   and tp_path_exists(irreps_layer_out_prev, edge_attr_irreps, ir)
            ]
        )

        irreps_layer_out = (irreps_scalars + irreps_gated).simplify()

        if nonlinearity_type == "gate":
            ir = (
                "0e"
                if tp_path_exists(irreps_layer_out_prev, edge_attr_irreps, "0e")
                else "0o"
            )
            irreps_gates = o3.Irreps([(mul, ir) for mul, _ in irreps_gated])

            # TO DO, it's not that safe to directly use the
            # dictionary
            equivariant_nonlin = Gate(
                irreps_scalars=irreps_scalars,
                act_scalars=[
                    acts[nonlinearity_scalars[ir.p]] for _, ir in irreps_scalars
                ],
                irreps_gates=irreps_gates,
                act_gates=[acts[nonlinearity_gates[ir.p]] for _, ir in irreps_gates],
                irreps_gated=irreps_gated,
            )

            conv_irreps_out = equivariant_nonlin.irreps_in.simplify()

        else:
            conv_irreps_out = irreps_layer_out.simplify()

            equivariant_nonlin = NormActivation(
                irreps_in=conv_irreps_out,
                # norm is an even scalar, so use nonlinearity_scalars[1]
                scalar_nonlinearity=acts[nonlinearity_scalars[1]],
                normalize=True,
                epsilon=1e-8,
                bias=False,
            )

        self.equivariant_nonlin = equivariant_nonlin

        # TODO: partial resnet?
        if irreps_layer_out == irreps_layer_out_prev and resnet:
            # We are doing resnet updates and can for this layer
            self.resnet = True
        else:
            self.resnet = False

        # TODO: last convolution should go to explicit irreps out
        logging.debug(
            f" parameters used to initialize {convolution.__name__}={convolution_kwargs}"
        )

        # override defaults for irreps:
        convolution_kwargs.pop("irreps_in", None)
        convolution_kwargs.pop("irreps_out", None)
        self.conv = convolution(
            irreps_in=self.irreps_in,
            irreps_out=conv_irreps_out,
            **convolution_kwargs,
        )

        # The output features are whatever we got in
        # updated with whatever the convolution outputs (which is a full graph module)
        self.irreps_out.update(self.conv.irreps_out)
        # but with the features updated by the nonlinearity
        self.irreps_out[
            AtomicDataDict.NODE_FEATURES_KEY
        ] = self.equivariant_nonlin.irreps_out

    def forward(self, data: ST) -> ST:
        # save old features for resnet
        old_x = data[AtomicDataDict.NODE_FEATURES_KEY]
        # run convolution
        data = self.conv(data)
        # do nonlinearity
        data[AtomicDataDict.NODE_FEATURES_KEY] = self.equivariant_nonlin(
            data[AtomicDataDict.NODE_FEATURES_KEY]
        )
        # do resnet
        if self.resnet:
            data[AtomicDataDict.NODE_FEATURES_KEY] = (
                    old_x + data[AtomicDataDict.NODE_FEATURES_KEY]
            )
        return data
